﻿using LocationSpecies.Domain;
using NHibernate;
using SqlWeb.Data;
using System;
using System.Data.Common;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using NHConfig = NHibernate.Cfg.Configuration;
using NHEnv = NHibernate.Cfg.Environment;

namespace WeatherPrep
{
    public class SessionHelper : IDisposable
    {
        static SessionHelper _instance;
        private ISessionFactory _sf;
        private bool _disposed;

        private SessionHelper()
        {
            _disposed = false;
            _sf = null;
        }

        /// <summary>
        /// Creates or retrieves the static instance of the SessionHelper
        /// </summary>
        /// <returns>The static instance of the SessionHelper.</returns>
        public static SessionHelper GetInstance()
        {
            SessionHelper instance = _instance;

            if (instance == null)
            {
                instance = new SessionHelper();

                // Set _instance to instance if null and return original value of _instance
                instance = Interlocked.CompareExchange(ref _instance, instance, null);

                // If instance is null, then _instance was null and set above.
                if (instance == null)
                {
                    _instance.CreateSessionFactory();
                }
            }

            return _instance;
        }

        /// <summary>
        /// NHibernate <see cref="NHibernate.ISessionFactory"/> for the i-Tree Data database.
        /// </summary>
        public ISessionFactory SessionFactory
        {
            get
            {
                return _sf;
            }
        }

        /// <summary>
        /// Open a session assocated with the i-Tree Data database.
        /// </summary>
        /// <returns>An <see cref="NHibernate.ISession"/> associated with the i-Tree Data database.</returns>
        public static ISession OpenSession()
        {
            SessionHelper instance = GetInstance();

            return instance.SessionFactory.OpenSession();
        }

        /// <summary>
        /// Initializes the SessionFactory of the current instance.
        /// </summary>
        private void CreateSessionFactory()
        {
            if (_sf == null || _sf.IsClosed)
            {
                Assembly asmLS = typeof(Location).Assembly;
                NHConfig config = GetNHConfig(asmLS);

                if (config == null)
                {
                    Assembly asm = Assembly.GetExecutingAssembly();
                    config = new NHConfig().Configure(asm, "WeatherPrep.LocationSpecies.cfg.xml");

#if DEBUG_SQL
                    config.Properties[NHEnv.ShowSql] = "true";
#endif

                    UpdateNHConfig(asmLS, config);
                }

                VerifyDatabaseVersion(config, asmLS);

                _sf = config.BuildSessionFactory();

                ValidateCache(config);
            }
        }

        /// <summary>
        /// Generates the path to where to store/load an NHibernate <see cref="NHibernate.Cfg.Configuration"/>.
        /// </summary>
        /// <param name="lib">domain library</param>
        /// <returns>The path to the configuration file.</returns>
        private string GetConfigFile(Assembly lib)
        {
            Assembly nhAssembly = typeof(NHConfig).Assembly;

            FileVersionInfo vNH = FileVersionInfo.GetVersionInfo(nhAssembly.Location);
            string appDir = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
            string cfgDir = Path.Combine(appDir, "i-Tree", "WeatherPrep");
            string cfgFile = String.Format(@"{0}.{1}.cfg", lib.GetName().Name, vNH.FileVersion);

            return Path.Combine(cfgDir, cfgFile);
        }

        /// <summary>
        /// Checks the validity of the configuration file against the domain library.
        /// </summary>
        /// <param name="cfgFile">configuration file</param>
        /// <param name="lib">domain library</param>
        /// <returns>true if the configuration file valid, false otherwise</returns>
        private bool ConfigValid(string cfgFile, Assembly lib)
        {
            if (!File.Exists(cfgFile))
            {
                return false;
            }

            FileInfo cfgInfo = new FileInfo(cfgFile);
            FileInfo libInfo = new FileInfo(lib.Location);

            return libInfo.LastWriteTime == cfgInfo.LastWriteTime;
        }

        /// <summary>
        /// Attempts to load the configuration file for the specified domain library.
        /// </summary>
        /// <param name="lib">domain library</param>
        /// <returns>NHibernate <see cref="NHibernate.Cfg.Configuration"/> if the configuration file is valid, null otherwise</returns>
        private NHConfig GetNHConfig(Assembly lib)
        {
            string cfgFile = GetConfigFile(lib);
            NHConfig config = null;

            if (ConfigValid(cfgFile, lib))
            {
                try
                {
                    using (FileStream fs = new FileStream(cfgFile, FileMode.Open))
                    {
                        BinaryFormatter bf = new BinaryFormatter();

                        config = bf.Deserialize(fs) as NHConfig;
                    }
                }
                catch (SerializationException)
                {
                    config = null;
                }
            }

            return config;
        }

        /// <summary>
        /// Updates the <see cref="NHibernate.Cfg.Configuration"/> for the domain libary
        /// </summary>
        /// <param name="lib">domain library</param>
        /// <param name="config">configuration</param>
        private void UpdateNHConfig(Assembly lib, NHConfig config)
        {
            string cfgFile = GetConfigFile(lib);
            string cfgDir = Path.GetDirectoryName(cfgFile);

            if (!Directory.Exists(cfgDir))
            {
                Directory.CreateDirectory(cfgDir);
            }

            using (FileStream fs = new FileStream(cfgFile, FileMode.Create))
            {
                BinaryFormatter bf = new BinaryFormatter();

                bf.Serialize(fs, config);
            }

            FileInfo libInfo = new FileInfo(lib.Location);
            File.SetLastWriteTime(cfgFile, libInfo.LastWriteTime);
        }

        /// <summary>
        /// Ensures that the domain version matches the version of database specified in the configuration.
        /// </summary>
        /// <param name="config">configuration</param>
        /// <param name="assembly">domain library</param>
        private void VerifyDatabaseVersion(NHConfig config, Assembly assembly)
        {
            Version databaseVersion = GetDatabaseVersion(config);
            Version domainVersion = GetAssemblyProductVersion(assembly);

            // If the major numbers do not match, then the domain library is not compatible
            // with the database version. Update the LocationSpecies.Domain package.
            if (databaseVersion.Major != domainVersion.Major)
            {
                throw new Exception("i-Tree Data version is incompatible with domain version.");
            }
        }

        /// <summary>
        /// Queries the database specified in the configuration to obtain the version number.
        /// </summary>
        /// <param name="config">configuration</param>
        /// <returns>The database version number</returns>
        private Version GetDatabaseVersion(NHConfig config)
        {
            string connectionString = config.Properties[NHEnv.ConnectionString];

            //cte 202202 Refactoring Needed for SqlWebConnection function, which goes to namespace SqlWeb.Data, shown below
            //// C:\iTree\WeatherPrep\WeatherPrep\packages\SqlWeb.Data.1.0.1-alpha-6\lib\net461\SqlWeb.Data.dll
            // Decompiled with ICSharpCode.Decompiler 6.1.0.5902
            /*
            public SqlWebConnection()
            {
                _builder = new SqlWebConnectionStringBuilder();
                _client = new RestClient();
                _client.ClearHandlers();
                _client.UseNewtonsoftJson();
                _state = ConnectionState.Closed;
            }
            */
            //cte 202202 Refactor to remove dependence on RestSharp.RestClient.ClearHandlers function, which was pulled out of the library
            //jk 202202 For developers looking to edit or update the WeatherPrep code: the packages "RestSharp" and "RestSharp.Serializers.NewtonsoftJson" cannot currently be updated beyond version 106.12. More recent updates are missing the needed RestSharp.RestClient.ClearHandlers function.
            using (DbConnection con = new SqlWebConnection(connectionString))
            using (DbCommand cmd = new SqlWebCommand())
            {
                con.Open();

                cmd.Connection = con;
                cmd.CommandText = "Select r.Value from _DatabaseRegistry r Where RegistryId = 4";

                object value = cmd.ExecuteScalar();

                if (value != null && value != DBNull.Value)
                {
                    return new Version(Convert.ToString(value));
                }
            }

            return null;
        }

        /// <summary>
        /// Obtains the product version number of the domain library which
        /// indicates the compatible database version number.
        /// </summary>
        /// <param name="assembly">domain library</param>
        /// <returns>The product version of the domain library</returns>
        private Version GetAssemblyProductVersion(Assembly assembly)
        {
            FileVersionInfo versionInfo = FileVersionInfo.GetVersionInfo(assembly.Location);

            return new Version(versionInfo.ProductVersion);
        }

        /// <summary>
        /// Clears the cache if the database version does not match the cached version.
        /// </summary>
        /// <param name="config"></param>
        private void ValidateCache(NHConfig config)
        {
            Version cacheVersion = GetCacheVersion();
            Version dbVersion = GetDatabaseVersion(config);

            if (dbVersion != cacheVersion)
            {
                ClearCache();

                // Ensure the new version is cached.
                GetCacheVersion();
            }
        }

        /// <summary>
        /// Retrieves the database version from the second level cache.
        /// </summary>
        /// <returns>The cached database version</returns>
        private Version GetCacheVersion()
        {
            Version version = null;

            using (ISession s = _sf.OpenSession())
            using (ITransaction tx = s.BeginTransaction())
            {
                RegistryEntry entry = s.Get<RegistryEntry>(4);

                if (!String.IsNullOrEmpty(entry.Value))
                {
                    version = new Version(entry.Value);
                }

                tx.Commit();
            }
            return version;
        }

        /// <summary>
        /// Clears all data from the cache.
        /// </summary>
        private void ClearCache()
        {
            var entities = _sf.GetAllClassMetadata().Keys;

            foreach (var entity in entities)
            {
                _sf.EvictEntity(entity);
            }

            // Clear the collection regions
            var roles = _sf.GetAllCollectionMetadata().Keys;

            foreach (var role in roles)
            {
                _sf.EvictCollection(role);
            }

            // Clear the default query region
            _sf.EvictQueries();
        }

        public void Dispose()
        {
            this.Dispose(true);

            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if (_disposed)
            {
                return;
            }

            if (disposing)
            {
                if (_sf != null && !_sf.IsClosed)
                {
                    _sf.Dispose();
                    _sf = null;
                }
            }

            Interlocked.CompareExchange(ref _instance, null, this);

            _disposed = true;
        }
    }
}
