﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;

namespace WeatherPrep
{
    /// <summary>
    /// Several functions to read raw surface weather data and compute weather variables
    /// </summary>
    public enum VEGTYPE { TREE, SHRUB, GRASS, CORN };

    public class SurfaceWeather
    {
        #region class_variables
        const double TEMP_CK = 273.15;                      ///<value>Conversino factor for Celsius to Kelvin</value>
        const double XLIM = 0.000000001;
        const double ABS_PRS = 1013.25;                     ///<value>mBar</value>
        const double ABS_TEMP = 288.15;                     ///<value>Kelvin</value>
        const int SOL_CONST = 1367;
        const double ES_CONS = 0.6108;                      ///<value>saturated vapor pressure constant (kPa)</value>
        const double ATRN = 0.957;
        const double K1 = 0.1;
        const double BA = 0.84;
        const double PA_CONS = 3.486;                       ///<value>Density for moist air sensible heat flux constant</value>
        const double CHI_CONS = 2.501;                      ///<value>latent heaf for evaoration of water, adjusted by temperature (used in meteo and evap routines)</value>
        const double GAMMA_C = 0.0016286;                   ///<value>psychrometric constant, adjusted by vapor pressure (used in meteo and evap routines)</value>
        const double theta = 0.0000000567;                  ///<value>Stefan-Boltzmann constant (MJ m^-2 K^-4 day^-1)</value>
        const double ES = 0.97;                             ///<value>surface emissivity for long wave radiation</value>
        const double ES_RN = 0.98;                          ///<value>surface emissivity (M.C. Llasat, R.L. Snyder, Ag and Forest Meteorology, 1998)</value>
        const double CP = 1013;                             ///<value>specific heat of moist air (J/Kg C)</value>
        const int ZCONS = 2;                                ///<value>height of weather station (m)</value>
        const double rd = 0.00137;                          ///<value>roughness height for water (m)</value>
        const double rDS = 0.005;                           ///<value>roughness height for snow (m)</value>
        const double rDT = 0.95;                            ///<value>roughness height for tree (m)</value>
        const double rTreeHeight = 5;                       ///<value>tree height (m)</value>
        const double rGroundHeight_m = 0.1;                  ///<value>ground height (m)</value>
        const double rZom = 0.123;                          ///<value>roughness length coefficient controlling momentum transfer defined Eq 13.3 Chin (2021)</value>
        const double rZov = 0.0123;                         ///<value>roughness length coefficient controlling heat and vapor transfer defined Eq 13.4 and following text Chin (2021)</value>
        const double rZu = ZCONS;                           ///<value>wind measurement height (m) from constants</value>
        const double rZe = ZCONS;                           ///<value>vapor measurement height (m) from constants</value>
        const double rZuT = rTreeHeight + rZu;              ///<value>wind estimate height in tree (m)</value>
        const double rZuG = 0.1 + rd;                       ///<value>wind estimate height for ground (m)</value>
        const double rZuG_m = rGroundHeight_m + rZu;                       ///<value>wind estimate height for ground (m)</value>
        double Height_WeatherStationWindSensor_m;
        double Height_TreeCanopyTop_m;

        public const double LEAF_STORAGE_M = 0.0002;               ///<value>specific leaf storage of water = 0.2 mm = 0.0002 m</value>
        const double IMPERV_STORAGE_M = 0.0015;             ///<value>maximum impervious cover storage of water = 1.5 mm = 0.0015 m</value>
        const double PERV_STORAGE_M = 0.001;                ///<value>maximum pervious cover storage of water = 1 mm = 0.001 m</value>                                    

        public const double NODATA = -999;
        const int NOWBAN = 99999;
        public const string SFC_TABLE = "SurfaceWeather";
        const string HRMIX_TABLE = "HourlyMixHt";

        const string NEW_INTERNATIONAL = "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW ZZ ZZ ZZ W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD";
        const string NEW_US_CANADA = "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB MW MW MW MW AW AW AW AW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD";
        const string OLD = "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD";
        const string NARCCAP = "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01      NETRAD PCPXX SD";
        // Field number for the old surface weather format
        const int USAF = 0;
        const int WBAN = 1;
        const int DATETIME = 2;
        const int DIR = 3;
        const int SPD = 4;
        const int GUS = 5;
        const int CLG = 6;
        const int SKC = 7;
        const int TEMP = 16;
        const int DEWP = 17;
        const int SLP = 18;
        const int ALT = 19;
        const int STP = 20;
        const int MAX = 21;
        const int MIN = 22;
        const int PCP01 = 23;
        const int NETRAD = 24;
        const int SD = 27;

        public DateTime TimeStamp;
        public int Jday;
        public int Year;
        public int Month;
        public int Day;
        public int Hour;
        public int Minute;
        public double DownLwWm2;
        public double UpLwWm2;
        public double LatRad;
        public double HourAngleRad;
        public double DecAngleRad;
        public double AirDens_kg_p_m3;
        public double AirMass;
        public double Altimeter;
        public double Ceiling;
        public double DewTempC;
        public double DewTempF;
        public double DifRadWm2;
        public double DirRadWm2;
        public double GlbRadWm2;
        public double NetRadWm2;
        public double OpCldCov;
        public double PrsIn;
        public double PrsKpa;
        public double PrsMbar;
        public double PeGrMh;
        public double PeTrMh;
        public double PeSnGrMh;
        public double PeSnTrMh;
        public double PtTrMh;
        public double VegEvMh;                  ///<value>Evaporation from vegetations (m/h)</value>
        public double VegStMh;                  ///<value>Storage of precipitation by vegetation (m/h)</value>
        public double VegIntcptMh;              ///<value>Precipitation interception by vegetation (m/h)</value>
        public double UnderCanThrufallMh;       ///<value>Under canopy throughfall of precipitation (m/h)</value>
        public double UnderCanPervEvMh;         ///<value>Evaporation from under canopy pervious cover (m/h)</value>
        public double UnderCanPervStMh;         ///<value>Storage of precipitation by under canopy pervious cover (m/h)</value>                                       
        public double UnderCanPervInfilMh;      ///<value>Infiltration from under canopy pervious cover (m/h)</value>
        public double UnderCanImpervEvMh;       ///<value>Evaporation from under canopy impervious cover (m/h)</value>
        public double UnderCanImpervStMh;       ///<value>Storage of precipitation by under canopy impervious cover (m/h)</value>                                       
        public double UnderCanImpervRunoffMh;   ///<value>Runoff from under canopy impervious cover (m/h)</value>
        public double NoCanPervEvMh;            ///<value>Evaporation from no canopy pervious cover (m/h)</value>
        public double NoCanPervStMh;            ///<value>Storage of precipitation by no canopy pervious cover (m/h)</value>                                       
        public double NoCanPervInfilMh;         ///<value>Infiltration from no canopy pervious cover (m/h)</value>
        public double NoCanImpervEvMh;          ///<value>Evaporation from no canopy impervious cover (m/h)</value>
        public double NoCanImpervStMh;          ///<value>Storage of precipitation by no canopy impervious cover (m/h)</value>                                       
        public double NoCanImpervRunoffMh;      ///<value>Runoff from no canopy impervious cover (m/h)</value>
        public double PARuEm2s;
        public double PARWm2;
        public double RainInH;
        public double RainIn6H;
        public double RainMh;
        public double RelHum;
        public double SatVPrsKpa;
        public double SnowIn;
        public double SnowM;
        public double SolZenAgl;
        public double TempC;
        public double TempF;
        public double TempK;
        public double ToCldCov;
        public double TrCldCov;
        public double VapPrsKpa;
        public double WdDir;
        public double WdSpdKnt;
        public double WdSpdMh;
        public double WdSpdMs;

        public enum WeatherDataFormat { OLD, NEWINTL, NEWUSCAN, NARCCAP };
        public static WeatherDataFormat weaFormat;
        #endregion

        //ProcessSurfaceWeatherData receives variables from CreateOutput.cs
        //MeteorologicalDataInputFile_string, PrecipitationDataInputFile_string, LocationData_Object, LeafAreaIndex_dataBase, VegType, startYear, endYear, WeatherVariables_dataBase
        public static void ProcessSurfaceWeatherData(string MeteorologicalDataInputFile_string, string PrecipitationDataInputFile_string, LocationData LocationData_Object,
            string LeafAreaIndex_dataBase, VEGTYPE VegetationType_class, int startYear, int endYear, string WeatherVariables_dataBase,
            double Height_WeatherStationWindSensor_m, double Height_TreeCanopyTop_m)
        {
            List<SurfaceWeather> rawList = new List<SurfaceWeather>();
            List<SurfaceWeather> intList = new List<SurfaceWeather>();
            List<SurfaceWeather> finList = new List<SurfaceWeather>();
            List<LeafAreaIndex> LeafAreaIndex_list = new List<LeafAreaIndex>();
            int recCnt = 0;

            try
            {
                ReadSurfaceWeatherData(MeteorologicalDataInputFile_string, ref rawList, startYear);
                CheckPressure(rawList, startYear, endYear);  //after this point Altimeter is no longer needed
                AdjustTimeStamp(rawList);
                recCnt = CreateHourlyRecords(rawList, ref intList, startYear, endYear);
                FillExtrapolate(intList);
                FillInterpolate(intList);
                ConvertData(intList, LocationData_Object.GMTOffset, recCnt);
                recCnt = ExtractValidData(intList, recCnt, ref finList);

                //If WeatherPrepConfig.xml has entry for PrecipitationDataCsv then update RainMh variable in finList 
                if (!string.IsNullOrEmpty(PrecipitationDataInputFile_string))
                {
                    ReplaceRainData(PrecipitationDataInputFile_string, ref finList);
                }

                CalcAirDensity(finList, recCnt);
                CalcSolarZenithAngle(finList, LocationData_Object, recCnt);
                CalcSolarRadiation(finList, LocationData_Object, recCnt);
                LeafAreaIndex.ReadLAIPartialRecords(LeafAreaIndex_dataBase, ref LeafAreaIndex_list, finList[0].TimeStamp, finList[recCnt - 1].TimeStamp);
                CalcET(finList, LeafAreaIndex_list, recCnt, Height_WeatherStationWindSensor_m, Height_TreeCanopyTop_m);
                CalcPrecipInterceptByCanopy(finList, LeafAreaIndex_list, recCnt, VegetationType_class);
                CalcPrecipInterceptByUnderCanopyCover(finList, recCnt);
                CalcPrecipInterceptByNoCanopyCover(finList, recCnt);
                //SurfaceWeather.CreateSurfaceWeatherTable(WeatherVariables_dataBase);
                SurfaceWeather.WriteSurfaceWeatherRecords(WeatherVariables_dataBase, finList, recCnt);
            }
            catch (Exception)
            {
                throw;
            }
        }
        //ReplaceRainData function uses XML-config provided rainFile to replace raw NOAA precipitation record
        static void ReplaceRainData(string rainFile, ref List<SurfaceWeather> list)
        {
            List<SurfaceWeather> rList = new List<SurfaceWeather>();
            int i, j;
            //Call ReadPrecip function by sending Precip.csv file and rList to store the precip data from the csv file
            ReadPrecip(rainFile, ref rList);

            // Search for the rain records that has the same time stamp as that for the first record in NOAA weather file.
            for (j = 0; j < rList.Count; j++)
            {
                if (list[0].TimeStamp == rList[j].TimeStamp) break;
            }
            for (i = 0; i < list.Count; i++)
            {
                if (list[i].TimeStamp == rList[j].TimeStamp)
                {
                    list[i].RainMh = rList[j++].RainMh;
                }
            }
            Console.WriteLine("Precipitation timeseries from PrecipitationDataFile replaced that from SurfaceWeatherDataFile.");
        }
        //ReadPrecip is used to parse Date, Time and Precipitation data from the optional Precip.csv input file for replacing NOAA raw weather values
        //JK 20211029: Renamed from "ReadCsv" to be more precise, and to avoid confusion with the functions handling the new CSV files for LAI & weather inputs
        static void ReadPrecip(string sFile, ref List<SurfaceWeather> rList)
        {
            string line;
            SurfaceWeather data;
            Regex reg;
            Match m;
            reg = new Regex(@"^\s*(?<dt>\d{8}),\s*(?<hr>\d{1,2}):(?<mn>\d{2}):(?<sc>\d{2}),\s*(?<p>[0-9.]+)\s*$");

            using (StreamReader sr = new StreamReader(sFile))
            {
                try
                {
                    //read the header line
                    sr.ReadLine();

                    while ((line = sr.ReadLine()) != null)
                    {
                        m = reg.Match(line);

                        data = new SurfaceWeather();

                        data.Year = int.Parse(m.Groups["dt"].ToString().Substring(0, 4));
                        data.Month = int.Parse(m.Groups["dt"].ToString().Substring(4, 2));
                        data.Day = int.Parse(m.Groups["dt"].ToString().Substring(6, 2));
                        data.Hour = int.Parse(m.Groups["hr"].ToString());
                        data.Minute = int.Parse(m.Groups["mn"].ToString());
                        data.TimeStamp = DateTime.Parse(data.Month + "/" +
                                                        data.Day + "/" +
                                                        data.Year + " " +
                                                        data.Hour + ":" +
                                                        data.Minute);
                        data.RainMh = double.Parse(m.Groups["p"].ToString());
                        rList.Add(data);
                    }
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }
        public static WeatherDataFormat CheckWeatherDataFormat(string sFile)
        {
            string line;
            WeatherDataFormat rc = WeatherDataFormat.OLD;

            using (StreamReader sr = new StreamReader(sFile))
            {
                try
                {
                    //read the header line
                    line = sr.ReadLine();
                    //Console.WriteLine("\nMeteorological Input Data Header: " + line + "\n");

                    if (Regex.IsMatch(line, "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW ZZ ZZ ZZ W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD"))
                    {
                        rc = WeatherDataFormat.NEWINTL;
                    }
                    else if (Regex.IsMatch(line, "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB MW MW MW MW AW AW AW AW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD"))
                    {
                        rc = WeatherDataFormat.NEWUSCAN;
                    }
                    else if (Regex.IsMatch(line, "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01 PCP06 PCP24 PCPXX SD"))
                    {
                        rc = WeatherDataFormat.OLD;
                    }
                    else if (Regex.IsMatch(line, "  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB WW WW WW W TEMP DEWP    SLP   ALT    STP MAX MIN PCP01      NETRAD PCPXX SD"))
                    {
                        rc = WeatherDataFormat.NARCCAP;
                    }
                    else
                    {
                        //throw new InvalidOperationException("\nWarning: WeatherPrepConfig.xml seems to point to a faulty SurfaceWeatherDataFile.");
                        Console.WriteLine("\nWarning: The WeatherPrepConfig.xml file contains a faulty value for SurfaceWeatherDataFile.");
                        Console.WriteLine("Check: " + sFile);
                        Console.WriteLine("Confirm: The header should contain  USAF  WBAN YR--MODAHRMN DIR SPD GUS CLG SKC L M H  VSB ...");
                        Console.WriteLine("Suggestion: Files downloaded from NOAA often need to be processed by ishapp2.exe.");
                    }
                    return rc;
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }
        //ReadSurfaceWeatherData function reads in the raw meteorological data, with USAF, WBAN, and other codes
        public static void ReadSurfaceWeatherData(string sFile, ref List<SurfaceWeather> WeatherVariables_list, int StartYear_input)
        {
            string line;
            SurfaceWeather data;
            Regex reg;

            weaFormat = CheckWeatherDataFormat(sFile);
            using (StreamReader sr = new StreamReader(sFile))
            {
                try
                {
                    switch (weaFormat)
                    {
                        case WeatherDataFormat.NEWINTL:
                            reg = new Regex(@"^(?<usaf>.{6}) (?<wban>.{5}) (?<dt>.{12}) (?<dir>.{3}) (?<spd>.{3}) .{3} (?<clg>.{3}) (?<skc>.{3}) .{30} (?<temp>.{4}) (?<dewp>.{4}) .{6} (?<alt>.{5}) (?<stp>.{6}) .{7} (?<pcp01>.{5}) (?<pcp06>.{5}) .{11} (?<sd>.{2})\s*$");
                            break;
                        case WeatherDataFormat.NEWUSCAN:
                            reg = new Regex(@"^(?<usaf>.{6}) (?<wban>.{5}) (?<dt>.{12}) (?<dir>.{3}) (?<spd>.{3}) .{3} (?<clg>.{3}) (?<skc>.{3}) .{36} (?<temp>.{4}) (?<dewp>.{4}) .{6} (?<alt>.{5}) (?<stp>.{6}) .{7} (?<pcp01>.{5}) (?<pcp06>.{5}) .{11} (?<sd>.{2})\s*$");
                            break;
                        case WeatherDataFormat.OLD:
                            reg = new Regex(@"^(?<usaf>.{6}) (?<wban>.{5}) (?<dt>.{12}) (?<dir>.{3}) (?<spd>.{3}) .{3} (?<clg>.{3}) (?<skc>.{3}) .{21} (?<temp>.{4}) (?<dewp>.{4}) .{6} (?<alt>.{5}) (?<stp>.{6}) .{7} (?<pcp01>.{5}) (?<pcp06>.{5}) .{11} (?<sd>.{2})\s*$");
                            break;
                        case WeatherDataFormat.NARCCAP:
                            reg = new Regex(@"^(?<usaf>.{6}) (?<wban>.{5}) (?<dt>.{12}) (?<dir>.{3}) (?<spd>.{3}) .{3} (?<clg>.{3}) (?<skc>.{3}) .{21} (?<temp>.{4}) (?<dewp>.{4}) .{6} (?<alt>.{5}) (?<stp>.{6}) .{7} (?<pcp>.{5}) (?<netrad>.{11}) .{5} (?<sd>.{2})\s*$");
                            break;
                        default:
                            reg = new Regex(@"^(?<usaf>.{6}) (?<wban>.{5}) (?<dt>.{12}) (?<dir>.{3}) (?<spd>.{3}) .{3} (?<clg>.{3}) (?<skc>.{3}) .{21} (?<temp>.{4}) (?<dewp>.{4}) .{6} (?<alt>.{5}) (?<stp>.{6}) .{7} (?<pcp01>.{5}) (?<pcp06>.{5}) .{11}  (?<sd>.{2})\s*$");
                            break;
                    }
                    Match m;
                    //read the header line
                    sr.ReadLine();

                    //CounterWhileLoop_int initialized to zero for error handling below, when CounterWhileLoop_int == 1
                    int CounterWhileLoop_int = 0;

                    while ((line = sr.ReadLine()) != null)
                    {
                        //CounterWhileLoop_int is advanced, and when CounterWhileLoop_int == 1 it will be used for error handling
                        CounterWhileLoop_int = CounterWhileLoop_int + 1;

                        line = Regex.Replace(line, "([0-9.]+)T", @"$1 ");   // remove trace (T) for PCP01
                        m = reg.Match(line);

                        data = new SurfaceWeather();

                        data.Year = (m.Groups["dt"].ToString().Substring(0, 4) == "****" || m.Groups["dt"].ToString().Substring(0, 4) == "    ") ? (int)NODATA : int.Parse(m.Groups["dt"].ToString().Substring(0, 4));

                        //If CounterWhileLoop_int equals 1 and the StartYear_input is not equal to data.Year then send error message
                        if (CounterWhileLoop_int == 1 & StartYear_input != data.Year)
                        {
                            //Notify user that the input instructions in the WeatherPrepConfigFile and weather data are problematic
                            Console.WriteLine("\nWarning: The WeatherPrepConfig.xml StartYear is " + StartYear_input + " and not the same as the input weather file year of " + data.Year + ".");
                            Console.WriteLine("Exiting: WeatherPrep cannot run when the desired year in WeatherPrepConfig.xml is different from the actual data.");
                            Console.WriteLine("Suggestion: Make sure the StartYear in WeatherPrepConfig.xml matches the year of your input data.");
                            //exit program when this error emerges
                            System.Environment.Exit(1);
                        }

                        data.Month = (m.Groups["dt"].ToString().Substring(4, 2) == "**" || m.Groups["dt"].ToString().Substring(4, 2) == "  ") ? (int)NODATA : int.Parse(m.Groups["dt"].ToString().Substring(4, 2));
                        data.Day = (m.Groups["dt"].ToString().Substring(6, 2) == "**" || m.Groups["dt"].ToString().Substring(6, 2) == "  ") ? (int)NODATA : int.Parse(m.Groups["dt"].ToString().Substring(6, 2));
                        data.Hour = (m.Groups["dt"].ToString().Substring(8, 2) == "**" || m.Groups["dt"].ToString().Substring(8, 2) == "  ") ? (int)NODATA : int.Parse(m.Groups["dt"].ToString().Substring(8, 2));
                        data.Minute = (m.Groups["dt"].ToString().Substring(10, 2) == "**" || m.Groups["dt"].ToString().Substring(10, 2) == "  ") ? (int)NODATA : int.Parse(m.Groups["dt"].ToString().Substring(10, 2));
                        data.TimeStamp = DateTime.Parse(data.Month + "/" +
                                                        data.Day + "/" +
                                                        data.Year + " " +
                                                        data.Hour + ":" +
                                                        data.Minute);
                        data.WdDir = (m.Groups["dir"].ToString() == "***" || m.Groups["dir"].ToString() == "   ") ? NODATA : double.Parse(m.Groups["dir"].ToString());
                        //read in wind spead (mph)
                        data.WdSpdMh = (m.Groups["spd"].ToString() == "***" || m.Groups["spd"].ToString() == "   ") ? NODATA : double.Parse(m.Groups["spd"].ToString());
                        //read in cloud ceiling (lowest opaque layer w/ 5/8 or greater coverage (hundreds ft); 722 = unlimited
                        data.Ceiling = (m.Groups["clg"].ToString() == "***" || m.Groups["clg"].ToString() == "   ") ? NODATA : double.Parse(m.Groups["clg"].ToString());
                        //read in sky cover (CLR-CLEAR, SCT-SCATTERED-1/8 TO 4/8, BKN-BROKEN-5/8 TO 7/8, OVC-OVERCAST, OBS-OBSCURED, POB-PARTIAL OBSCURATION
                        switch (m.Groups["skc"].ToString())
                        {
                            case "***":
                            case "   ":
                                data.ToCldCov = NODATA;
                                break;
                            case "CLR":
                                data.ToCldCov = 0;
                                break;
                            case "SCT":
                                data.ToCldCov = 3.75;
                                break;
                            case "BKN":
                                data.ToCldCov = 7.5;
                                break;
                            case "OVC":
                                data.ToCldCov = 10;
                                break;
                            default:
                                data.ToCldCov = 10;
                                break;
                        }
                        //read in temperature (F)
                        data.TempF = (m.Groups["temp"].ToString() == "****" || m.Groups["temp"].ToString() == "    ") ? NODATA : double.Parse(m.Groups["temp"].ToString());
                        //read in dew point (F)
                        data.DewTempF = (m.Groups["dewp"].ToString() == "****" || m.Groups["dewp"].ToString() == "    ") ? NODATA : double.Parse(m.Groups["dewp"].ToString());
                        //read altimeter setting (inches and hundredths Hg)
                        data.Altimeter = (m.Groups["alt"].ToString() == "*****" || m.Groups["alt"].ToString() == "     ") ? NODATA : double.Parse(m.Groups["alt"].ToString());
                        //read in station atmospheric pressure (mbar)
                        data.PrsMbar = (m.Groups["stp"].ToString() == "******" || m.Groups["stp"].ToString() == "      ") ? NODATA : double.Parse(m.Groups["stp"].ToString());
                        //read in liquid precipitation for 1-hr (inches)
                        data.RainInH = (m.Groups["pcp01"].ToString() == "*****" || m.Groups["pcp01"].ToString() == "     ") ? 0 : double.Parse(m.Groups["pcp01"].ToString());
                        if (weaFormat == WeatherDataFormat.NARCCAP)
                        {
                            //read in net radiation (W/m2)
                            data.NetRadWm2 = (m.Groups["netrad"].ToString() == "***********" || m.Groups["netrad"].ToString() == "           ") ? NODATA : double.Parse(m.Groups["netrad"].ToString());
                        }
                        else
                        {
                            //read in liquid precipitation for 6-hr (inches)
                            data.RainIn6H = (m.Groups["pcp06"].ToString() == "*****" || m.Groups["pcp06"].ToString() == "     ") ? 0 : double.Parse(m.Groups["pcp06"].ToString());
                        }
                        //read in snow depth (inches)
                        data.SnowIn = (m.Groups["sd"].ToString() == "**" || m.Groups["sd"].ToString() == "  ") ? 0 : double.Parse(m.Groups["sd"].ToString());
                        WeatherVariables_list.Add(data);
                    }
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }
        /// <summary>
        /// Check missing data count in pressure and altimeter setting. If the number of pressure data is not enough
        /// but altimeter data, convert altimeter setting to pressure.
        /// </summary>
        /// <remarks>Only weather stations with no more than 48 hours of consecutive missing data can be used for Eco analyses.
        /// Those with more than 48-hour consecutive missing data in both PrsMbar and Altimeter are excluded.
        /// Therefore, there could be a case in which PrsMbar is totally missing and Altimeter doesn't have 48-hour consecutive missing data.
        /// Here only this case is treated; if PrsMbar has more than 48-hour missing data and Altimeter doesn't,
        /// convert Altimeter to PrsMbar.</remarks>
        /// <param name="WeatherVariables_list">Surface weather data list</param>
        /// <param name="startYear">Start year for processing</param>
        /// <param name="endYear">End year for processing</param>
        public static void CheckPressure(List<SurfaceWeather> WeatherVariables_list, int startYear, int endYear)
        {
            if ((CheckPrsMbar(WeatherVariables_list, startYear, endYear) == false) &&
                (CheckAltimeter(WeatherVariables_list, startYear, endYear) == true))
            {
                Alt2Prs(WeatherVariables_list);
            }
        }
        /// <summary>
        /// Check if pressure variable is available or not.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        /// <param name="startYear">Start year for processing</param>
        /// <param name="endYear">End year for processing</param>
        /// <returns>True if pressure data is available, False if not</returns>
        static bool CheckPrsMbar(List<SurfaceWeather> list, int startYear, int endYear)
        {
            int i;
            DateTime lastDateTime;
            DateTime endDt;
            bool result = false;
            TimeSpan ts;

            //set true unless everything is NODATA
            for (i = 0; i < list.Count; i++)
            {
                if (list[i].PrsMbar != NODATA)
                {
                    result = true;
                    break;
                }
            }

            //check if the first measurement is more than 48 hrs apart from January 1 0:00
            lastDateTime = DateTime.Parse("1/1/" + startYear);
            ts = list[0].TimeStamp - lastDateTime;
            if (ts.TotalHours > 48)
            {
                result = false; ;
                return (result);
            }

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].PrsMbar != NODATA)
                {
                    lastDateTime = list[i].TimeStamp;
                }
                else
                {
                    ts = list[i].TimeStamp - lastDateTime;
                    if (ts.TotalHours > 48)
                    {
                        result = false;
                        return (result);
                    }
                }
            }

            //check if the last measurement is more than 48 hrs apart from 12/31/endYear 23:00
            endDt = DateTime.Parse("12/31/" + endYear + " 23:00");
            ts = endDt - lastDateTime;
            if (ts.TotalHours > 48)
            {
                result = false;
                return (result);
            }
            return (result);
        }
        /// <summary>
        /// Check if altimeter setting variable is available or not.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        /// <param name="startYear">Start year for processing</param>
        /// <param name="endYear">End year for processing</param>
        /// <returns>True if altimeter setting data is available, False if not</returns>
        static bool CheckAltimeter(List<SurfaceWeather> list, int startYear, int endYear)
        {
            int i;
            DateTime lastDateTime;
            DateTime endDt;
            bool result = false;
            TimeSpan ts;

            //set true unless everything is NODATA
            for (i = 0; i < list.Count; i++)
            {
                if (list[i].Altimeter != NODATA)
                {
                    result = true;
                    break;
                }
            }

            //check if the first measurement is more than 48 hrs apart from January 1 0:00
            lastDateTime = DateTime.Parse("1/1/" + startYear);
            ts = list[0].TimeStamp - lastDateTime;
            if (ts.TotalHours > 48)
            {
                result = false; ;
                return (result);
            }

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].Altimeter != NODATA)
                {
                    lastDateTime = list[i].TimeStamp;
                }
                else
                {
                    ts = list[i].TimeStamp - lastDateTime;
                    if (ts.TotalHours > 48)
                    {
                        result = false;
                        return (result);
                    }
                }
            }

            //check if the last measurement is more than 48 hrs apart from 12/31/endYear 23:00
            endDt = DateTime.Parse("12/31/" + endYear + " 23:00");
            ts = endDt - lastDateTime;
            if (ts.TotalHours > 48)
            {
                result = false;
                return (result);
            }
            return (result);
        }
        /// <summary>
        /// Convert altimeter setting values to pressure values.
        /// </summary>
        /// <param name="list">Surrace weather list</param>
        static void Alt2Prs(List<SurfaceWeather> list)
        {
            int i;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].Altimeter == NODATA)
                {
                    list[i].PrsMbar = NODATA;
                }
                else
                {
                    list[i].PrsMbar = list[i].Altimeter * 100 * 0.338639;
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of surface weather time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        public static void FillExtrapolate(List<SurfaceWeather> list)
        {
            ExtrapoloteCeiling(list);
            ExtrapolatePrsMbar(list);
            ExtrapolateToCldCov(list);
            ExtrapolateTempF(list);
            ExtrapolateDewTempF(list);
            ExtrapolateWdDir(list);
            ExtrapolateWdSpdMh(list);
            if (weaFormat == WeatherDataFormat.NARCCAP)
            {
                ExtrapolateNetRadWm2(list);
            }
        }
        /// <summary>
        /// Fill missing values at start or end of ceiling height time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapoloteCeiling(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].Ceiling == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].Ceiling != NODATA)
                    {
                        dblFill = list[i].Ceiling;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].Ceiling = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].Ceiling == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].Ceiling != NODATA)
                    {
                        dblFill = list[i].Ceiling;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].Ceiling = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of pressure time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolatePrsMbar(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].PrsMbar == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].PrsMbar != NODATA)
                    {
                        dblFill = list[i].PrsMbar;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].PrsMbar = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].PrsMbar == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].PrsMbar != NODATA)
                    {
                        dblFill = list[i].PrsMbar;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].PrsMbar = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of cloud cover time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateToCldCov(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].ToCldCov == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].ToCldCov != NODATA)
                    {
                        dblFill = list[i].ToCldCov;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].ToCldCov = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].ToCldCov == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].ToCldCov != NODATA)
                    {
                        dblFill = list[i].ToCldCov;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].ToCldCov = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of temperature time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateTempF(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].TempF == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].TempF != NODATA)
                    {
                        dblFill = list[i].TempF;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].TempF = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].TempF == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].TempF != NODATA)
                    {
                        dblFill = list[i].TempF;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].TempF = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of dew point temperature time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateDewTempF(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].DewTempF == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].DewTempF != NODATA)
                    {
                        dblFill = list[i].DewTempF;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].DewTempF = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].DewTempF == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].DewTempF != NODATA)
                    {
                        dblFill = list[i].DewTempF;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].DewTempF = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of wind direction time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateWdDir(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].WdDir == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].WdDir != NODATA)
                    {
                        dblFill = list[i].WdDir;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].WdDir = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].WdDir == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].WdDir != NODATA)
                    {
                        dblFill = list[i].WdDir;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].WdDir = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of wind speed time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateWdSpdMh(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].WdSpdMh == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].WdSpdMh != NODATA)
                    {
                        dblFill = list[i].WdSpdMh;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].WdSpdMh = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].WdSpdMh == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].WdSpdMh != NODATA)
                    {
                        dblFill = list[i].WdSpdMh;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].WdSpdMh = dblFill;
                    }
                }
            }
        }
        /// <summary>
        /// Fill missing values at start or end of net radiation time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void ExtrapolateNetRadWm2(List<SurfaceWeather> list)
        {
            //This process fills in the missing values in the beginning of a list
            int i, j;
            double dblFill = 0;

            if (list[0].NetRadWm2 == NODATA)
            {
                //looks for the value that is not NODATA
                i = 1;
                while (true)
                {
                    if (list[i].NetRadWm2 != NODATA)
                    {
                        dblFill = list[i].NetRadWm2;
                        break;
                    }
                    i++;
                    if (i == list.Count)
                    {
                        break;
                    }
                }
                if (i != list.Count)
                {
                    for (j = i - 1; j >= 0; j--)
                    {
                        list[j].NetRadWm2 = dblFill;
                    }
                }
            }
            if (list[list.Count - 1].NetRadWm2 == NODATA)
            {
                //looks for the value that is not NODATA
                i = list.Count - 1;
                while (true)
                {
                    if (list[i].NetRadWm2 != NODATA)
                    {
                        dblFill = list[i].NetRadWm2;
                        break;
                    }
                    i--;
                    if (i == 0)
                    {
                        break;
                    }
                }
                if (i != 0)
                {
                    for (j = i + 1; j < list.Count; j++)
                    {
                        list[j].NetRadWm2 = dblFill;
                    }
                }
            }
        }

        /// <summary>
        /// Fill missing values in the middle of surface weather time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        public static void FillInterpolate(List<SurfaceWeather> list)
        {
            InterpolatePrsMbar(list);
            InterpolateTempF(list);
            InterpolateDewTempF(list);
            InterpolateWdDir(list);
            InterpolateWdSpdMh(list);
            InterpolateCeiling(list);
            InterpolateToCldCov(list);
            if (weaFormat == WeatherDataFormat.NARCCAP)
            {
                InterpolateNetRadWm2(list);
            }
        }
        /// <summary>
        /// Fill missing values in the middle of pressure time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolatePrsMbar(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].PrsMbar == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].PrsMbar != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].PrsMbar - list[bfIdx].PrsMbar;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].PrsMbar = list[bfIdx].PrsMbar + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of temperature time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateTempF(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].TempF == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].TempF != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].TempF - list[bfIdx].TempF;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].TempF = list[bfIdx].TempF + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of dew point temperature time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateDewTempF(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].DewTempF == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].DewTempF != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].DewTempF - list[bfIdx].DewTempF;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].DewTempF = list[bfIdx].DewTempF + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of wind direction time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateWdDir(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].WdDir == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].WdDir != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].WdDir - list[bfIdx].WdDir;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].WdDir = list[bfIdx].WdDir + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of wind speed time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateWdSpdMh(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].WdSpdMh == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].WdSpdMh != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].WdSpdMh - list[bfIdx].WdSpdMh;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].WdSpdMh = list[bfIdx].WdSpdMh + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of ceiling height time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateCeiling(List<SurfaceWeather> list)
        {
            int i, j, k;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].Ceiling == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].Ceiling != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        list[k].Ceiling = list[afIdx].Ceiling;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Fill missing values in the middle of total cloud cover time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateToCldCov(List<SurfaceWeather> list)
        {
            int i, j, k;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].ToCldCov == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].ToCldCov != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        list[k].ToCldCov = list[afIdx].ToCldCov;
                    }
                    i = j;
                }
            }
        }
        /// Fill missing values in the middle of net radiation time series data if exists.
        /// </summary>
        /// <param name="list">Surface weather list</param>
        static void InterpolateNetRadWm2(List<SurfaceWeather> list)
        {
            int i, j, k;
            double valDif, stpInc;
            int timeDif;
            int bfIdx;      //index for the record before the records with missing values
            int afIdx;      //index for the record after the records with missing values
            TimeSpan ts;

            for (i = 0; i < list.Count; i++)
            {
                if (list[i].NetRadWm2 == NODATA)
                {
                    bfIdx = i - 1;
                    j = i + 1;
                    while (true)
                    {
                        if (list[j].NetRadWm2 != NODATA)
                        {
                            afIdx = j;
                            break;
                        }
                        else
                        {
                            j++;
                        }
                    }

                    //taking the before to after differences of time and values
                    ts = list[afIdx].TimeStamp - list[bfIdx].TimeStamp;
                    timeDif = (int)ts.TotalMinutes;
                    valDif = list[afIdx].NetRadWm2 - list[bfIdx].NetRadWm2;
                    //taking the increment value of the difference
                    stpInc = valDif / timeDif;
                    //looping through the NODATA value
                    for (k = bfIdx + 1; k < afIdx; k++)
                    {
                        ts = list[k].TimeStamp - list[bfIdx].TimeStamp;
                        list[k].NetRadWm2 = list[bfIdx].NetRadWm2 + stpInc * ts.TotalMinutes;
                    }
                    i = j;
                }
            }
        }
        /// <summary>
        /// Set timestamp to the nearest top-of-hour.
        /// </summary>
        /// <remarks> Below is a typical timestamp observed in ISH data. Data are typically recorded
        /// one per hour but the timestamp is not on the top-of-hour. During rain events, precipitation (PCP01)
        /// and other data tend to be recorded more often. This function adjusts the timestamp to the
        /// nearest next top-of-hour.
        /// 
        /// YR--MODAHRMN 
        /// 200501050054 -> 0100
        /// 200501050154 -> 0200
        /// 200501050254 -> 0300
        /// 200501050339 -> 0400
        /// 200501050354 -> 0400
        /// 200501050407 -> 0500
        /// 200501050454 -> 0500
        /// 200501050515 -> 0600
        /// 200501050536 -> 0600
        /// 200501050554 -> 0600
        /// 200501050600 -> 0600
        /// 200501050605 -> 0700
        /// 200501050623 -> 0700
        /// 200501050654 -> 0700
        /// 200501050700 -> 0700
        /// 
        /// Among those records with the same time stamp (e.g. 0600), variable values found
        /// the last will be used in the hourly records. For PCP01, the largest value will 
        /// be used.  </remarks>
        /// 
        /// <param name="rawList">Surface weather list</param>
        public static void AdjustTimeStamp(List<SurfaceWeather> rawList)
        {
            int i;

            for (i = 0; i < rawList.Count; i++)
            {
                if (rawList[i].TimeStamp.Minute != 0)
                {
                    rawList[i].TimeStamp = rawList[i].TimeStamp.AddMinutes(60 - rawList[i].TimeStamp.Minute);
                }
            }
        }
        /// <summary>
        /// With timestamps adjusted, rawList may have multiple records at the same hour.
        /// This method creates hourly time series data by extracting variable values
        /// found the last within the same timestamp records. For rain, the largest value
        /// within the same timestamp records will be extracted.
        /// </summary>
        /// <param name="rawList">List holding original surface weather data</param>
        /// <param name="hrList">List holding surface weather hourly data</param>
        /// <param name="startYear">Start year for processing</param>
        /// <param name="endYear">End year for processing</param>
        /// <returns>Record count in hrList list</returns>
        /// <summary>
        public static int CreateHourlyRecords(List<SurfaceWeather> rawList, ref List<SurfaceWeather> hrList,
                                                int startYear, int endYear)
        {
            DateTime startDt;
            int j;      //index for the raw data lists
            int i;      //index for the hourly data lists
            int recCnt = 0;
            SurfaceWeather data;

            for (i = startYear; i <= endYear; i++)
            {
                if (DateTime.IsLeapYear(i))
                {
                    recCnt += 24 * 366;
                }
                else
                {
                    recCnt += 24 * 365;
                }
            }
            startDt = DateTime.Parse("1/1/" + startYear + " 0:00");
            hrList = new List<SurfaceWeather>();

            // set the blank data with only timestamp throughout the year(s)
            //Loop through all records to set TimeStamp to incremental hours after start date
            //Loop through all records to set initial values of variables = NoData or = 0
            for (i = 0; i < recCnt; i++)
            {
                data = new SurfaceWeather();
                data.TimeStamp = startDt.AddHours(i);
                data.WdDir = data.WdSpdMh = data.Ceiling = data.ToCldCov =
                data.TempF = data.DewTempF = data.PrsMbar = data.NetRadWm2 = NODATA;
                data.RainInH = data.SnowIn = 0;
                hrList.Add(data);
            }
            try
            {
                j = 0;
                //Loop through all records to set certain variables to values != NoData if the raw data had non-zero values 
                //Loop through all records to set certain variables to values > 0 if the raw data had non-zero values 
                for (i = 0; i < recCnt; i++)
                {
                    //If the new and raw lsit TimeStamp are equal do the next comparison of values != NoData or >= 0
                    if (hrList[i].TimeStamp == rawList[j].TimeStamp)
                    {
                        do
                        {
                            if (rawList[j].RainInH >= hrList[i].RainInH)
                            {
                                hrList[i].RainInH = rawList[j].RainInH;
                            }
                            if (rawList[j].SnowIn >= hrList[i].SnowIn)
                            {
                                hrList[i].SnowIn = rawList[j].SnowIn;
                            }
                            if (rawList[j].WdDir != NODATA)
                            {
                                hrList[i].WdDir = rawList[j].WdDir;
                            }
                            if (rawList[j].WdSpdMh != NODATA)
                            {
                                hrList[i].WdSpdMh = rawList[j].WdSpdMh;
                            }
                            if (rawList[j].Ceiling != NODATA)
                            {
                                hrList[i].Ceiling = rawList[j].Ceiling;
                            }
                            if (rawList[j].ToCldCov != NODATA)
                            {
                                hrList[i].ToCldCov = rawList[j].ToCldCov;
                            }
                            if (rawList[j].TempF != NODATA)
                            {
                                hrList[i].TempF = rawList[j].TempF;
                            }
                            if (rawList[j].DewTempF != NODATA)
                            {
                                hrList[i].DewTempF = rawList[j].DewTempF;
                            }
                            if (rawList[j].PrsMbar != NODATA)
                            {
                                hrList[i].PrsMbar = rawList[j].PrsMbar;
                            }
                            if (weaFormat == WeatherDataFormat.NARCCAP)
                            {
                                if (rawList[j].NetRadWm2 != NODATA)
                                {
                                    hrList[i].NetRadWm2 = rawList[j].NetRadWm2;
                                }
                            }
                            j++;
                            if (j == rawList.Count)
                            {
                                return (recCnt);
                            }
                        } while (hrList[i].TimeStamp == rawList[j].TimeStamp);
                    }
                }
                return (recCnt);
            }
            catch (Exception)
            {
                return (recCnt);
                throw;
            }
        }
        /// <summary>
        /// Convert unit of surface weather to that required by i-Tree Eco.
        /// </summary>
        /// <param name="intList">List holding top of hour surface weather data</param>
        /// <param name="timeZone">Hour offset to GMT</param>
        /// <param name="recCnt">Record count in intList</param>
        public static void ConvertData(List<SurfaceWeather> intList, double timeZone, int recCnt)
        {
            DateTime dateTimeLST;
            int i;

            try
            {
                for (i = 0; i < recCnt; i++)
                {
                    dateTimeLST = intList[i].TimeStamp.AddHours((int)timeZone);    //GMT to LST

                    intList[i].TimeStamp = dateTimeLST;
                    intList[i].Year = dateTimeLST.Year;
                    intList[i].Month = dateTimeLST.Month;
                    intList[i].Day = dateTimeLST.Day;
                    intList[i].Hour = dateTimeLST.Hour;
                    intList[i].Jday = dateTimeLST.DayOfYear;
                    intList[i].PrsIn = intList[i].PrsMbar / 33.8639 * 100;       //mBar to hundredths of inches
                    intList[i].PrsKpa = intList[i].PrsMbar / 10;                 //mBar to kPa
                    intList[i].RainMh = intList[i].RainInH * 0.0254;             //inches/hour to meters/hour
                    intList[i].SnowM = intList[i].SnowIn * 0.0254;               //inches to meters
                    intList[i].TempC = (intList[i].TempF - 32) * 5 / 9;        //degrees Fahrenheit to degrees Celsius
                    intList[i].TempK = intList[i].TempC + TEMP_CK;               //degrees Celsius to Kelvin
                    intList[i].DewTempC = (intList[i].DewTempF - 32) * 5 / 9;  //degrees Fahrenheit to degrees Celsius
                    intList[i].WdSpdKnt = Math.Round(intList[i].WdSpdMh / 1.151);      //Miles/hour to knot
                    intList[i].WdSpdMs = intList[i].WdSpdMh / 2.237;             //Miles/hour to meters/sec
                }
            }
            catch (Exception)
            {
                throw;
            }
        }
        /// <summary>
        /// Extract data for the analysis year from intermediate top of hour data (intList) 
        /// </summary>
        /// <remarks>Original surface weather data are recorded with UTC timestamp. These data were converted to the 
        /// local time by ConvertData method, which may create data for previous or next year at the beginning or
        /// ending of the data, depending on time offset from GMT. This method eleminates data for previous or next 
        /// year and extract data for the analysis year only.</remarks>
        /// <param name="intList">List holding intermediate top of hour data</param>
        /// <param name="recCnt">Record count in intList</param>
        /// <param name="finList">List holding extracted top of hour data</param>
        /// <returns>Record count in finList</returns>
        public static int ExtractValidData(List<SurfaceWeather> intList, int recCnt, ref List<SurfaceWeather> finList)
        {
            int i;
            int startIdx = 0;
            int endIdx = 0;
            int finIdx;
            SurfaceWeather data;

            //Due to GMT offset, the records don't start at 0:00 or end at 23:00
            if (intList[0].Hour != 0 || intList[recCnt - 1].Hour != 23)
            {
                //Search the first record with the hour 0:00
                for (i = 0; i < recCnt; i++)
                {
                    if (intList[i].Hour == 0)
                    {
                        startIdx = i;
                        break;
                    }
                }
                //Search the last record with the hour 23:00
                for (i = recCnt - 1; i >= 0; i--)
                {
                    if (intList[i].Hour == 23)
                    {
                        endIdx = i;
                        break;
                    }
                }
                recCnt = endIdx - startIdx + 1;
                finList = new List<SurfaceWeather>();

                for (i = 0; i < recCnt; i++)
                {
                    data = new SurfaceWeather();
                    finList.Add(data);
                }

                finIdx = 0;
                for (i = startIdx; i <= endIdx; i++)
                {
                    finList[finIdx].TimeStamp = intList[i].TimeStamp;
                    finList[finIdx].Jday = intList[i].Jday;
                    finList[finIdx].Year = intList[i].Year;
                    finList[finIdx].Month = intList[i].Month;
                    finList[finIdx].Day = intList[i].Day;
                    finList[finIdx].Hour = intList[i].Hour;

                    finList[finIdx].Ceiling = intList[i].Ceiling;
                    finList[finIdx].DewTempF = intList[i].DewTempF;
                    finList[finIdx].RainInH = intList[i].RainInH;
                    finList[finIdx].PrsMbar = intList[i].PrsMbar;
                    finList[finIdx].SnowIn = intList[i].SnowIn;
                    finList[finIdx].TempF = intList[i].TempF;
                    finList[finIdx].ToCldCov = intList[i].ToCldCov;
                    finList[finIdx].WdDir = intList[i].WdDir;
                    finList[finIdx].WdSpdMh = intList[i].WdSpdMh;

                    finList[finIdx].PrsIn = intList[i].PrsIn;
                    finList[finIdx].PrsKpa = intList[i].PrsKpa;
                    finList[finIdx].RainMh = intList[i].RainMh;
                    finList[finIdx].SnowM = intList[i].SnowM;
                    finList[finIdx].TempC = intList[i].TempC;
                    finList[finIdx].TempK = intList[i].TempK;
                    finList[finIdx].DewTempC = intList[i].DewTempC;
                    finList[finIdx].WdSpdKnt = intList[i].WdSpdKnt;
                    finList[finIdx].WdSpdMs = intList[i].WdSpdMs;

                    if (weaFormat == WeatherDataFormat.NARCCAP)
                    {
                        finList[finIdx].NetRadWm2 = intList[i].NetRadWm2;
                    }
                    finIdx++;
                }
            }
            else
            {
                finList = intList;
            }
            return (recCnt);
        }
        /// <summary>
        /// Calculate air density.
        /// </summary>
        /// <param name="WeatherVariables_list">List holding hourly surface weather data</param>
        /// <param name="recCnt">Record count in WeatherVariables_list</param>
        public static void CalcAirDensity(List<SurfaceWeather> WeatherVariables_list, int recCnt)
        {
            int i;

            for (i = 0; i < recCnt; i++)
            {
                //DensityAir_kg_p_m3 (kg/m3) is density of air from Eq 4.2.4 from Shuttleworth (1993), w correction in conversion to C from K
                //Note: Given equation uses TempK, the 273.15 in the denominator is removed
                //Note: Eq 4.24 used incorrect denominator of 275 rather than 273.15 to convert from C to K; see Chin (2021) Eq 13.51
                //Note: Chin Eq 13.51 which unfortunately uses 3.45 in place of the correct 3.486.
                //Note: This tested well against values of air density from the EngineeringToolbox.com for air temperature from 0 to 50 C at standard atmospheric pressure
                WeatherVariables_list[i].AirDens_kg_p_m3 = 3.486 * (WeatherVariables_list[i].PrsKpa / WeatherVariables_list[i].TempK);
            }
        }
        /// <summary>
        /// Calculate solar zenith angle.
        /// </summary>
        /// <param name="WeatherVariables_list">List holding hourly surface weather data</param>
        /// <param name="LocationData_Object">Location data</param>
        /// <param name="recCnt">Record count in WeatherVariables_list</param>
        public static void CalcSolarZenithAngle(List<SurfaceWeather> WeatherVariables_list, LocationData LocationData_Object, int recCnt)
        {
            double dayAngle;  //day angle
            double solDecl; //solar declination
            double eqnTime;   //equation of time
            double stdMer;    //standard meridian
            double lat;       //local latitude in radians
            double trSolTime; //true solar time in hours
            double hourAngle; //hour angle
            double cosZ;      //cosine of the zenith angle
            double exAtmZenAgl;  //exoatmospheric solar zenith angle
            double exAtmElvAgl; //exoatmospheric solar elevation angle
            double refCorr;   //refraction correction
            int i;

            for (i = 0; i < recCnt; i++)
            {
                //dayAngle (radians)
                dayAngle = Deg2Rad(360 * (WeatherVariables_list[i].Jday - 1) / 365.0);
                //eqnTime (decimal minutes), equation of time
                eqnTime = (0.000075 + 0.001868 * Math.Cos(dayAngle) - 0.032077 * Math.Sin(dayAngle)
                        - 0.014615 * Math.Cos(2 * dayAngle) - 0.040849 * Math.Sin(2 * dayAngle)) * 229.18;
                //solDecl (radians), solar declination angle
                solDecl = (0.006918 - 0.399912 * Math.Cos(dayAngle) + 0.070257 * Math.Sin(dayAngle)
                            - 0.006758 * Math.Cos(2 * dayAngle) + 0.000907 * Math.Sin(2 * dayAngle)
                            - 0.002697 * Math.Cos(3 * dayAngle) + 0.00148 * Math.Sin(3 * dayAngle));
                //WeatherVariables_list stores solDecl (radians)
                WeatherVariables_list[i].DecAngleRad = solDecl;
                //To handle both Northern/Southern hemisphere, accept negative latitude (for Southern)
                //lat (radians), latitude
                lat = Deg2Rad(LocationData_Object.Latitude);
                //WeatherVariables_list stores latitude
                WeatherVariables_list[i].LatRad = lat;
                //stdMer (degrees) takes 15 deg for each GMT offset hour
                stdMer = 15.0 * LocationData_Object.GMTOffset;
                //trSolTime (hour with decimal minutes), true solar time
                trSolTime = WeatherVariables_list[i].Hour + (4 * (Math.Abs(stdMer) - Math.Abs(LocationData_Object.Longitude)) + eqnTime) / 60.0;
                //hourAngle (degrees)
                hourAngle = 15.0 * trSolTime - 180;
                //hourAngle (radians)
                hourAngle = Deg2Rad(hourAngle);
                //WeatherVariables_list stores hourAngle
                WeatherVariables_list[i].HourAngleRad = hourAngle;
                //cosZ (radians) cosine Zenith angle
                cosZ = Math.Sin(solDecl) * Math.Sin(lat) + Math.Cos(solDecl) * Math.Cos(lat) * Math.Cos(hourAngle);
                if (cosZ > 1)
                {
                    cosZ = 1.0;
                }
                else if (cosZ < -1)
                {
                    cosZ = -1.0;
                }

                //calculate atmosphere correction for the solar zenith angle
                exAtmZenAgl = Rad2Deg(Math.Acos(cosZ));    //in degrees
                exAtmElvAgl = 90.0 - exAtmZenAgl;           //in degrees

                refCorr = 0.0;
                if (-1 <= exAtmElvAgl && exAtmElvAgl < 15)
                {
                    refCorr = ABS_PRS / ABS_TEMP
                            * (0.1594 + 0.0196 * exAtmElvAgl + 0.00002 * Math.Pow(exAtmElvAgl, 2))
                            / (1 + 0.505 * exAtmElvAgl + 0.0845 * Math.Pow(exAtmElvAgl, 2));
                }
                else if (15 <= exAtmElvAgl && exAtmElvAgl < 90)
                {
                    refCorr = 0.00452 * ABS_PRS / ABS_TEMP / Math.Tan(Deg2Rad(exAtmElvAgl));
                }
                WeatherVariables_list[i].SolZenAgl = exAtmZenAgl - refCorr;
            }
        }
        /// <summary>
        /// Convert unit of angle from degrees to radians.
        /// </summary>
        /// <param name="angle">Angle in degrees</param>
        /// <returns>Angle in radians</returns>
        static double Deg2Rad(double angle)
        {
            return Math.PI * angle / 180.0;
        }
        /// <summary>
        /// Convert unit of angle from radians to degrees.
        /// </summary>
        /// <param name="angle">Angle in radians</param>
        /// <returns>Angle in degrees</returns>
        static double Rad2Deg(double angle)
        {
            return 180.0 * angle / Math.PI;
        }
        /// <summary>
        /// Calculate solar radiation.
        /// </summary>
        /// <param name="WeatherVariables_list">List holding hourly surface weather data</param>
        /// <param name="LocationData_Object">Location data</param>
        /// <param name="recCnt">Record count in WeatherVariables_list</param>
        public static void CalcSolarRadiation(List<SurfaceWeather> WeatherVariables_list, LocationData LocationData_Object, int recCnt)
        {
            double ETR;       //ETR:   extraterrestrial radiation
            double ETRN;      //ETRN:
            double pcAirMass; //M:     pressure-corrected relative optical air mass
            double opqn;      //OPQN:  adjusted opaque cloud cover for diffuse radiation calculation
            double opqd;      //OPQD:  adjusted opaque cloud cover for direct radiation calculation
            double Kn;        //Kn:    transmission value of extraterrestrial direct normal radiation
            double Tr;        //TR:    Rayleigh scattering transmittance
            double Toz;       //TO:    ozone absorption transmittance
            double Tum;       //TUM:   uniformly mixed gas transmittance
            double Ta;        //TA:    aerosol absorption and scattering transmittance
            double Kd;
            int i;

            for (i = 0; i < recCnt; i++)
            {
                CalcExoatmRad(WeatherVariables_list[i].Jday, WeatherVariables_list[i].SolZenAgl, out ETR, out ETRN);

                CalcAirMass(WeatherVariables_list[i].SolZenAgl, ref WeatherVariables_list[i].AirMass, WeatherVariables_list[i].PrsKpa, out pcAirMass);
                CalcCloudCover(WeatherVariables_list[i].AirMass, WeatherVariables_list[i].Ceiling, WeatherVariables_list[i].ToCldCov,
                                ref WeatherVariables_list[i].OpCldCov, ref WeatherVariables_list[i].TrCldCov, out opqn, out opqd);
                CalcTransmissionValue(WeatherVariables_list[i].Day, WeatherVariables_list[i].AirMass, WeatherVariables_list[i].TempC, WeatherVariables_list[i].DewTempC,
                                WeatherVariables_list[i].PrsMbar, ref WeatherVariables_list[i].SatVPrsKpa, ref WeatherVariables_list[i].VapPrsKpa, ref WeatherVariables_list[i].RelHum,
                                LocationData_Object.ozone[WeatherVariables_list[i].Month - 1], LocationData_Object.CoeffA, LocationData_Object.CoeffPHI, LocationData_Object.CoeffC,
                                LocationData_Object.Elevation, pcAirMass, opqn, out Tr, out Toz, out Tum, out Ta, out Kn);
                CalcDiffuseValue(WeatherVariables_list[i].AirMass, WeatherVariables_list[i].OpCldCov, WeatherVariables_list[i].TrCldCov, WeatherVariables_list[i].RainMh,
                                LocationData_Object.AlbedoVal[WeatherVariables_list[i].Month - 1], Tr, Toz, Tum, Ta, opqd, Kn, out Kd);

                WeatherVariables_list[i].DirRadWm2 = Kn * ETRN;
                WeatherVariables_list[i].DifRadWm2 = Kd * ETR;
                WeatherVariables_list[i].GlbRadWm2 = (Kn + Kd) * ETR;
                WeatherVariables_list[i].PARWm2 = WeatherVariables_list[i].GlbRadWm2 * 0.46;
                WeatherVariables_list[i].PARuEm2s = WeatherVariables_list[i].PARWm2 * 4.6;

                if (weaFormat != WeatherDataFormat.NARCCAP)
                {
                    CalcNetRad(WeatherVariables_list[i].TempK, WeatherVariables_list[i].DewTempC, WeatherVariables_list[i].ToCldCov,
                                WeatherVariables_list[i].GlbRadWm2, ref WeatherVariables_list[i].DownLwWm2, ref WeatherVariables_list[i].UpLwWm2, ref WeatherVariables_list[i].NetRadWm2, LocationData_Object.AlbedoVal[WeatherVariables_list[i].Month - 1]);
                }
            }
        }
        /// <summary>
        /// Calculate 
        /// </summary>
        /// <param name="jDay"></param>
        /// <param name="zenAgl"></param>
        /// <param name="ETR"></param>
        /// <param name="ETRN"></param>
        static void CalcExoatmRad(int jDay, double zenAgl, out double ETR, out double ETRN)
        {
            double dayAngle;  //psi:   day angle
            double eccent;    //eo:    eccentricity of the earth's orbit

            dayAngle = Deg2Rad(360 * (jDay - 1) / 365.0);                        //in radians
            eccent = 1.00011 + 0.034221 * Math.Cos(dayAngle) + 0.00128 * Math.Sin(dayAngle)
                    + 0.000719 * Math.Cos(2 * dayAngle) + 0.000077 * Math.Sin(2 * dayAngle);
            ETR = eccent * SOL_CONST * Math.Cos(Deg2Rad(zenAgl));
            ETRN = eccent * SOL_CONST;

            if (ETR < 0)
            {
                ETR = 0.0;
                ETRN = 0.0;
            }
        }
        static void CalcAirMass(double zenAgl, ref double airMass, double prsKpa, out double pcAirMass)
        {
            double solElvAgl; //alpha: solar elevation angle

            solElvAgl = 90.0 - zenAgl;   //in degrees
            if (zenAgl <= 90)
            {
                airMass = 1 / (Math.Sin(Deg2Rad(solElvAgl)) + 0.50572 * Math.Pow((solElvAgl + 6.07995), -1.6364));
            }
            else
            {
                airMass = 99.0;
            }
            if (airMass >= 30)
            {
                airMass = 30.0;
            }
            pcAirMass = airMass * (prsKpa / 10.0) / 1013.0;
        }
        static void CalcCloudCover(double airMass, double Ceiling, double total,
                            ref double opaque, ref double trans, out double opqn, out double opqd)
        {
            double a1;
            double b1;
            double corrN;
            double corrD;

            if (Ceiling > 700 && total <= 5)
            {
                opaque = total;
            }
            else if (Ceiling > 700 && total > 5)
            {
                opaque = 5.0;
            }
            else if (Ceiling <= 70 && total >= 0)
            {
                opaque = total;
            }
            else
            {
                opaque = 0.0;
            }
            trans = total - opaque;

            if (1 <= opaque && opaque <= 9)
            {
                a1 = 4.955 * (1 - Math.Exp(-0.454 * airMass)) - 3.4;
                if (a1 <= 0)
                {
                    b1 = -0.2 * a1;
                }
                else
                {
                    b1 = 0.1 * a1;
                }
                corrN = a1 * Math.Sin(Math.PI / 10 * total) + b1 * Math.Sin(2 * Math.PI / 10 * total);
                corrD = a1 * Math.Sin(Math.PI / 10 * total) * 0.5;
            }
            else
            {
                corrN = 0.0;
                corrD = 0.0;
            }
            opqn = opaque + corrN;
            opqd = opaque + corrD;
        }
        static void CalcTransmissionValue(int day, double airMass, double tempC, double dewC, double prsMb,
                            ref double SatVPrsKpa, ref double VapPrsKpa, ref double RelHum, double ozone, double a,
                            double phi, double C, double elev, double pcAirMass, double opqn, out double Tr,
                            out double Toz, out double Tum, out double Ta, out double Kn)
        {
            double Tw;        //Tw:    water vapor absorption transmittance
            double Ttrn;      //TTRN:  translucent cloud transmittance
            double Topq;      //TOPQ:  opaque cloud cover transmittance
            double Xo;        //XO:    total amount of ozone in cm in a slanted path
            double tau;       //tauA:  broadband aerosol optical depth
            double vPrsMb;    //vapor pressure in mBar
            double vPrsMbAdj; //pressure adjusted vapor pressure in mBar
            double Uw;        //Uw:    total precipitable water
            double Xw;        //Xw:    precipitable water vapor in cm in a slant path

            //Rayleigh scattering
            Tr = Math.Exp((-0.0903 * Math.Pow(pcAirMass, 0.84)) * (1 + pcAirMass - Math.Pow(pcAirMass, 1.01)));
            Xo = ozone * airMass;
            Toz = 1 - 0.1611 * Xo * Math.Pow((1 + 139.45 * Xo), -0.3035)
                - 0.002715 * Xo / (1 + 0.044 * Xo + 0.0003 * Math.Pow(Xo, 2));
            Tum = Math.Exp(-0.0127 * Math.Pow(pcAirMass, 0.26));
            tau = a * Math.Sin(360 * (double)day / 365 - phi) + C;
            Ta = Math.Exp(-tau * airMass);

            SatVPrsKpa = ES_CONS * Math.Exp((17.27 * tempC) / (237.3 + tempC));
            VapPrsKpa = ES_CONS * Math.Exp((17.27 * dewC) / (237.3 + dewC));
            RelHum = VapPrsKpa / SatVPrsKpa;
            vPrsMb = VapPrsKpa * 10.0;
            vPrsMbAdj = vPrsMb * prsMb / 1013.25;
            Uw = vPrsMbAdj * (0.0004 * elev + 1.1) + 1;   //in millimeter
            Uw = Uw / 10.0; //in cm
            Xw = Uw * airMass;
            Tw = 1 - 1.668 * Xw / (Math.Pow((1 + 54.6 * Xw), 0.637) + 4.042 * Xw);

            //Translucent cloud transmittance
            //    if weatherData.TrCldCov[i]  > 0 {
            //        amTrn = airMass
            //        if amTrn >= 10# {
            //            amTrn = 10#
            //        }
            //        bTrn = 0.012 + 0.006 * weatherData.OpCldCov[i] 
            //        Ttrn = ATRN - bTrn * amTrn
            //    }
            Ttrn = 1;

            //Opaque cloud
            Topq = (10 - opqn) / 10.0;

            Kn = 0.9751 * Tr * Toz * Tum * Tw * Ta * Topq * Ttrn;
        }
        static void CalcDiffuseValue(double airMass, double opaque, double Trans, double rain, double albedo, double
                        Tr, double Toz, double Tum, double Ta, double opqd, double Kn, out double Kd)
        {
            double Ksr;       //KSR:   diffuse radiation from Rayleigh scattering
            double Tas;       //TAS:   trensmittance of aerosol scattering
            double Taa;       //TAA:   transmittance of aerosol absorptance
            double Ksa;       //KSA:   diffuse radiation from aerosol scattering
            double fM;        //f(M):  empirical air mass function
            double Ksopq;     //KSOPQ:
            double Kstrn;     //KSTRN:
            double Ksgrf;     //KSGRF: diffuse radiation from ground reflectance
            double Ksgrf1;    //KSGRF1:
            double Ksgrf2;    //KSGRF2:

            double b2;
            double c2;
            double psw;       //PSW:
            double Kd0;
            double Rcld;    //RCLD:  broadband (solar) cloud reflectance
            double Ratm;      //RATM:  broadband (solar) atmospheric reflectance

            Taa = 1 - K1 * (1 - airMass + Math.Pow(airMass, 1.06)) * (1 - Ta);
            Ksr = 0.5 * (1 - Tr) * Toz * Tum * Taa;
            Ksa = BA * (1 - Ta) * Toz * Tum * Taa;
            fM = 0.38 + 0.925 * Math.Exp(-0.851 * airMass);

            //    if trans > 0 {
            //        Kstrn = -0.00235 + 0.00689 * trans + 0.000203 * trans ^ 2
            //    }
            Kstrn = 0.0;

            Ksopq = 0.0;
            if (opqd >= 5)
            {
                b2 = 0.0953 + 0.137 * opqd - 0.0409 * Math.Pow(opqd, 2)
                    + 0.00579 * Math.Pow(opqd, 3) - 0.000328 * Math.Pow(opqd, 4);

                c2 = -0.109 - 0.02 * opqd + 0.011 * Math.Pow(opqd, 2)
                    - 0.00156 * Math.Pow(opqd, 3) + 0.000121 * Math.Pow(opqd, 4);

                Ksopq = -0.06 + b2 * Ta + c2 * Math.Pow(Ta, 2);

                if (Ksopq <= 0)
                {
                    Ksopq = 0.0;
                }
            }

            if (opaque >= 8)
            {
                if (rain != 0)
                {
                    psw = 0.6;
                }
                else
                {
                    psw = 1.0;
                }
            }
            else
            {
                psw = 1.0;
            }

            Kd0 = (fM * (Ksr + Ksa) + Ksopq + Kstrn) * psw;

            Rcld = 0.06 * opaque + 0.02 * Trans;
            Ksgrf1 = (Kn + Kd0) * Rcld * (albedo - 0.2);
            if (Ksgrf1 <= 0)
            {
                Ksgrf1 = 0.01;
            }
            Tas = Ta / Taa;
            Ratm = (0.0685 + 0.16 * (1 - Tas)) * (10 - opaque) / 10.0;
            Ksgrf2 = (Kn + Kd0) * Ratm * albedo;
            Ksgrf = Ksgrf1 + Ksgrf2;

            Kd = Kd0 + Ksgrf;
        }
        static void CalcNetRad(double TempK, double dewC, double ToCldCov, double
                            glbRad, ref double downLw, ref double upLw, ref double netRad, double albedo)
        {
            //Reference:
            //Hirabayashi, S., & Endreny, T. (2016). Surface and Upper Weather Pre-processor for i-Tree Eco and Hydro. i-Tree Tools White Paper, Version 1.2. 

            double emissivity;
            double sfcLongRad;    //Long wave radiation from the surface (W/m^2)
            double cldLongRad;    //Long wave radiaton from cloud cover w/ Emissivity set to 1 (W/m^2)
            double skyLongRad;    //Long wave radiation from clear sky with Emissivity set to Ea*(1-cc) (W/m^2)
            double totShtRad;     //Total albedo corrected incoming short wave radiation (W/m^2)
            double totLongRad;    //Total long wave radiation emissivity corrected (incoming sky - outgoing surface) (W/m^2)
            double emissivity_of_1 = 1.0;

            //Emissivity of a clear sky (fraction), Eq 58 Hirabayashi and Endreny (2016)
            emissivity = 0.741 + (0.0062 * dewC);
            //Long wave radiation from the surface (W/m^2), Eq 59 Hirabayashi and Endreny (2016)
            sfcLongRad = theta * Math.Pow(TempK, 4);
            //Long wave radiaton downwelling from cloud cover (W/m2) w/ Emissivity set to 1 (W/m^2), Eq 57 Hirabayashi and Endreny (2016)
            cldLongRad = emissivity_of_1 * ToCldCov / 10.0 * theta * Math.Pow(TempK, 4);
            //Long wave radiation from clear sky (W/m2) with Emissivity set to Ea*(1-cc) (W/m^2), Eq 56 Hirabayashi and Endreny (2016)
            skyLongRad = emissivity * (1 - ToCldCov / 10) * theta * Math.Pow(TempK, 4);

            //Total albedo corrected incoming short wave radiation (W/m^2), Eq 54 Hirabayashi and Endreny (2016)
            totShtRad = (1 - albedo) * glbRad;
            //Total long wave radiation emissivity corrected (incoming sky - outgoing surface) (W/m^2), Eq 55 Hirabayashi and Endreny (2016)
            totLongRad = ES * (skyLongRad + cldLongRad - sfcLongRad);
            //Total net radation (downwelling short wave and long wave) (W/m^2), Eq 53 Hirabayashi and Endreny (2016)
            netRad = totShtRad + totLongRad;

            //downLw and upLw sent back to function calling CalcNetRad for use in Radiation.csv
            //Total longwave radiation downward (W/m2), Part of Eq 55 Hirabayashi and Endreny (2016)
            downLw = ES * (skyLongRad + cldLongRad);
            //Total longwave radiation upward (W/m2), Part of Eq 55 Hirabayashi and Endreny (2016)
            upLw = ES * sfcLongRad;

        }
        //Function CalcET using descriptive variable names with units
        public static void CalcET(List<SurfaceWeather> weatherData, List<LeafAreaIndex> LeafAreaIndex_list, int recCnt,
            double Height_WeatherStationWindSensor_m, double Height_TreeCanopyTop_m)
        {
            //References
            //Chin, D. A. (2021). Water Resources Engineering, Fourth Edition. Hoboken, NJ: Pearson Education.
            //Fassnacht, S.R. 2004 Estimating..Snowpack Simulation, Hydrologic Processes 18:3481-3492 https://onlinelibrary.wiley.com/doi/10.1002/hyp.5806
            //Jensen, M. E., & Allen, R. G. (2015). Evaporation, Evapotranspiration, and Irrigation Water Requirements: Manual of Practice 70, Second Edition (2nd ed.). Reston, VA: American Society of Civil Engineers.
            //Light, P. (1941). Analysis of high rates of snow-melting. Eos, Transactions American Geophysical Union, 22(1), 195-205. doi:https://doi.org/10.1029/TR022i001p00195
            //Lundberg, A., Calder, I., & Harding, R. (1998). Evaporation of intercepted snow: measurement and modelling. Journal of Hydrology, 206(3-4), 151-163. 
            //McCutcheon, S. C., Martin, J. L., & Barnwell, T. O. J. (1993). Water Quality. In D. R. Maidment (Ed.), Handbook of Hydrology (pp. 11.11-11.73). New York: McGraw-Hill.
            //Shuttleworth, J. W. (1993). Evaporation. In D. R. Maidment (Ed.), Handbook of Hydrology (pp. 4.1-4.5.3). New York: McGraw-Hill.
            //Stull, R. B. (2000). Meteorology for Scientists and Engineers (2nd ed.). Pacific Grove, CA: Brooks/Cole.
            //UCAR (2011).Snowmelt Processes: International Edition. University Corporation for Atmospheric Research COMET MetEd Program. https://download.comet.ucar.edu/memory-stick/hydro/basic_int/snowmelt/index.htm

            int i;
            double VaporPressure_Actual_kPa;
            double VaporPressure_Saturated_kPa;
            double VaporPressure_Deficit_kPa;
            double VaporPressure_Deficit_Pa;
            double VaporPressure_Deficit_OverWater_Pa;
            double WindSpeed_Station_m_p_s;
            double WindSpeed_TreeCanopy_m_p_s;
            double DensityWater_kg_p_m3;
            double DensityAir_kg_p_m3;
            double Radiation_Net_Wpm2;
            double LeafAreaIndex_Active_Tree_m2_p_m2;
            double ResistanceSurface_Tree_and_Soil_s_p_m;
            double ResistanceSurface_Water_s_p_m;
            double ResistanceAerodynamic_TreeCanopy_s_p_m;
            double ResistanceAerodynamic_SurfaceWater_s_p_m;
            double ResistanceAerodynamic_SnowTreeCanopy_s_p_m;
            double ResistanceAerodynamic_SnowGround_s_p_m;
            double vonKarman_Constant = 0.41;
            double Ratio_sec_to_hr = 3600.0;
            double Ratio_Pa_to_kPa = 1000.0;
            double Ratio_J_to_MJ = 1000000.0;
            double Ratio_MJ_to_J = 1.0 / 1000000.0;
            //Roughness Lengths from Davenport-Wieringa Zo (m) to Classification and Landscape Table 4.1 in Stull (2000)
            //Note: Classificaiton smooth = Snow; Classificaiton Open = Airport; Note UCAR Community Land Model uses 0.024 
            double RoughnessLength_Snow_m = 0.005;
            double RoughnessLength_Airport_m = 0.03;
            //Roughness Lengths for open water from Shuttleworth (1993) page 4.15, above Eq 4.2.30, citing Thom and Oliver (1977)
            double RoughnessLength_Water_m = 0.00137;
            //ZeroPlaneDisplacementHeight_frac (fraction) from Eq 13.3 Chin (2021) adjusts cover height to get zero plane displacement height
            double ZeroPlaneDisplacementHeight_frac = 0.67;
            //Height_TreeWindSensor_m created to avoid negative logarithms below, ensuring tree height is below wind sensor height
            //Note: Error Handling: If users enter wind sensor height = 2 m, tree height = 5 m, then aerodynamic resistance equations fail
            //Note: ... Aerodynamic resistance equations take log of difference of sensor height and zero plane displacement height of object
            //Note: ... tree height * zero plane displacement fraction = 5 * 0.67 = 3.35; ln(2-3.35) = No Solution
            double Height_TreeWindSensor_m = Height_WeatherStationWindSensor_m + Height_TreeCanopyTop_m;
            double Height_TreeZeroPlaneDisplacement_m = Height_TreeCanopyTop_m * ZeroPlaneDisplacementHeight_frac;
            //Height_Ground_m is set to 0.1 m to represent surface irregularity
            //Note: If Height_Ground_m converted to roughness length with Zom scale factor = 0.123 then it has classification between open and smooth 
            double Height_Ground_m = 0.1;
            double Height_GroundWindSensor_m = Height_WeatherStationWindSensor_m + Height_Ground_m;
            double Height_GroundZeroPlaneDisplacement_m = Height_Ground_m * ZeroPlaneDisplacementHeight_frac;
            double LAI_plus_BAI_plus_GAI_Tree_to_SVeg_ratio;
            double LAI_plus_BAI_plus_GAI_Ground_to_Tree_ratio;
            //BarkAreaIndex_Tree_m2_p_m2 for trees is estimated at 1.7 from HydroPlus test cases
            double BarkAreaIndex_Tree_m2_p_m2 = 1.7;
            //LeafAreaIndex_SVeg_m2_p_m2 for shrubs is estimated at 2.3 from HydroPlus test cases
            double LeafAreaIndex_SVeg_m2_p_m2 = 2.2;
            //BarkAreaIndex_SVeg_m2_p_m2 for shrubs is estimated at 0.5 from HydroPlus test cases
            double BarkAreaIndex_SVeg_m2_p_m2 = 0.5;
            //GroundAreaIndex_m2_p_m2 for trees, shrubs, and ground is estimated at 1
            double GroundAreaIndex_m2_p_m2 = 1;

            //SpecificHeat_MoistAir_J_p_kg_p_C is specific heat of moist air at constant pressure, defined by Chin (2021) w Eq 13.37
            double SpecificHeat_MoistAir_J_p_kg_p_C = 1013;
            double LatentHeatOfVaporizaton_J_p_kg;
            double LatentHeatOfVaporizaton_MJ_p_kg;
            double PsychrometricConstant_Pa_p_C;
            double PsychrometricConstant_kPa_p_C;
            double VaporPressureGradient_Pa;
            double VaporPressureGradient_kPa;
            double Ratio_MolecularWeightVapor_to_MolecularWeightDryAir = 0.622;
            double SpecificHeat_MoistAir_MJ_p_kg_p_C;
            double GroundHeatFlux_W_p_m2;
            double ResistanceSurface_Vegetation_s_p_m;
            double BulkStomatalResistance_Adjustment_frac;
            double HeatStorage_Water_J_p_m2_p_s;
            double ResistanceAerodynamic_Snow_to_Rain_ratio;


            try
            {
                //For loop through recCnt, the length of the input data
                for (i = 0; i < recCnt; i++)
                {
                    //VaporPressure_Actual_kPa is actual vapor pressure (kPa)
                    VaporPressure_Actual_kPa = 0.6108 * Math.Exp((17.27 * weatherData[i].DewTempC) / (237.3 + weatherData[i].DewTempC));
                    VaporPressure_Saturated_kPa = 0.6108 * Math.Exp((17.27 * weatherData[i].TempC) / (237.3 + weatherData[i].TempC));
                    //VaporPressure_Deficit_kPa is vapor pressure deficit (kPa), using previously computed saturated vapor pressure SatVPrsKpa (kPa)
                    VaporPressure_Deficit_kPa = (VaporPressure_Saturated_kPa - VaporPressure_Actual_kPa);
                    //VaporPressure_Deficit_Pa (Pa) is vapor pressure deficit in fundamental units used in Penman-Monteith Eq 13.1 of Chin (2021) 
                    VaporPressure_Deficit_Pa = VaporPressure_Deficit_kPa * Ratio_Pa_to_kPa;
                    //VaporPressure_Deficit_OverWater_Pa (Pa) is vapor pressure deficit over water, initially equal to at station value
                    VaporPressure_Deficit_OverWater_Pa = VaporPressure_Deficit_Pa;
                    //Note: 202212 This option currently remains off until further evidence that it is needed.
                    //Note: Adjustment to approximate theory of Jensen and Allen (2015) Chapter 6, Evaporation from Water Surfaces:
                    //Note: ... "Evaporation can be low during summer when shortwave radiative energy is absorbed by the cold water bodies and
                    //Note: ... when vapor pressure gradients above the water are small due to high humidity of air caused by regional ET.
                    //Note: ... Evaporation is high during winter as dry air from nontranspiring frozen regions is coupled with saturation
                    //Note: ... vapor pressure at the water surface that may be much higher than the air."
                    //If TempC > freezing then reduce vapor pressure deficit due to high humidity of air by evaporation; otherwise no adjustment
                    /*
                    if (weatherData[i].TempC > 0) {
                        //VaporPressure_Deficit_Reduction_frac set to 0.75, but should be adjusted to fit evidence of altered deficit
                        double VaporPressure_Deficit_Reduction_frac = 0.75;
                        //VaporPressure_Deficit_Adjustment_frac (frac) formula derived here; should be adjusted to fit observations; 50 degrees used as asymptote
                        double VaporPressure_Deficit_Adjustment_frac = 1 - (Math.Pow(weatherData[i].TempC / 50, 2) * VaporPressure_Deficit_Reduction_frac);
                        //If VaporPressure_Deficit_Adjustment_frac goes below VaporPressure_Deficit_Reduction_frac, then set to that limit
                        if (VaporPressure_Deficit_Adjustment_frac < VaporPressure_Deficit_Reduction_frac) {
                            VaporPressure_Deficit_Adjustment_frac = VaporPressure_Deficit_Reduction_frac; }
                        //VaporPressure_Deficit_OverWater_Pa (Pa) reduced to VaporPressure_Deficit_Adjustment_frac its magnitude
                        VaporPressure_Deficit_OverWater_Pa = VaporPressure_Deficit_OverWater_Pa * VaporPressure_Deficit_Adjustment_frac;
                    }
                    */

                    //WindSpeed_Station_m_p_s is wind speed at station (m/s)
                    WindSpeed_Station_m_p_s = weatherData[i].WdSpdMs;
                    //If WindSpeed_Station_m_p_s < = zero, then set to 0.1 m/s to avoid division by zero
                    if (WindSpeed_Station_m_p_s <= 0)
                    {
                        WindSpeed_Station_m_p_s = 0.1;
                    }
                    //WindSpeed_TreeCanopy_m_p_s defined as wind speed at tree top (m/s) using Eq 4.14b of Stull (2000)
                    WindSpeed_TreeCanopy_m_p_s = WindSpeed_Station_m_p_s * (Math.Log(Height_TreeWindSensor_m / RoughnessLength_Airport_m) /
                        Math.Log(Height_WeatherStationWindSensor_m / RoughnessLength_Airport_m));

                    //DensityWater_kg_p_m3 is density of water (kg/m^3) from Equation in Figure 11.1.1 McCutcheon (1993)
                    DensityWater_kg_p_m3 = 1000 * (1 - ((weatherData[i].TempC + 288.9414) / (508929.2 * (weatherData[i].TempC + 68.12963))) *
                        Math.Pow((weatherData[i].TempC - 3.9863), 2));

                    //DensityAir_kg_p_m3 (kg/m3) is density of air from Eq 4.2.4 from Shuttleworth (1993), w correction in conversion to C from K
                    //Note: Eq 4.24 used incorrect denominator of 275 rather than 273.15 to convert from C to K; see Chin (2021) Eq 13.51
                    //Note: Chin Eq 13.51 which unfortunately uses 3.45 in place of the correct 3.486.
                    //Note: This tested well against values of air density from the EngineeringToolbox.com for air temperature from 0 to 50 C at standard atmospheric pressure
                    DensityAir_kg_p_m3 = 3.486 * (weatherData[i].PrsKpa / (273.15 + weatherData[i].TempC));

                    //LatentHeatOfVaporizaton_MJ_p_kg (MJ/kg) from Eq 13.36 Chin (2021) is latent heat of vaporization
                    LatentHeatOfVaporizaton_MJ_p_kg = 2.501 - 0.002361 * weatherData[i].TempC;
                    //LatentHeatOfVaporizaton_J_p_kg (J/kg) is converted from MJ to K
                    LatentHeatOfVaporizaton_J_p_kg = LatentHeatOfVaporizaton_MJ_p_kg * Ratio_J_to_MJ;

                    //PsychrometricConstant_kPa_p_C (kPa/C) from Eq 13.37 Chin (2021) is psychrometric constant
                    //Note: Eq 13.37 uses specific heat with units of kJ/kg/C
                    SpecificHeat_MoistAir_MJ_p_kg_p_C = SpecificHeat_MoistAir_J_p_kg_p_C * Ratio_MJ_to_J;
                    PsychrometricConstant_kPa_p_C = (SpecificHeat_MoistAir_MJ_p_kg_p_C * weatherData[i].PrsKpa) /
                        (Ratio_MolecularWeightVapor_to_MolecularWeightDryAir * LatentHeatOfVaporizaton_MJ_p_kg);
                    //PsychrometricConstant_Pa_p_C (Pa/C) is converted from kPa to Pa
                    PsychrometricConstant_Pa_p_C = PsychrometricConstant_kPa_p_C * Ratio_Pa_to_kPa;

                    //VaporPressureGradient_kPa (kPa) from Eq 13.43 Chin (2021) is gradient of saturation pressure vs temperature curve
                    VaporPressureGradient_kPa = 4098 * (0.6108 * Math.Exp((17.27 * weatherData[i].TempC) / (weatherData[i].TempC + 237.3))) /
                        Math.Pow((weatherData[i].TempC + 237.3), 2);
                    //VaporPressureGradient_Pa (Pa) is converted from kPa to Pa
                    VaporPressureGradient_Pa = VaporPressureGradient_kPa * Ratio_Pa_to_kPa;

                    //Radiation_Net_Wpm2 (W/m2) is energy from downwelling shortwave and longwave radiation (W/m^2)
                    Radiation_Net_Wpm2 = weatherData[i].NetRadWm2;

                    //GroundHeatFlux_W_p_m2 (W/m2) defined for two vegetation heights after Eq 13.32 by Chin (2021); also Jensen et. al., (2015) Chapter 5
                    //Mote: Jensen and Allen (2015) "Standardized Reference Evapotranspiration for Short Reference ETos: Reference ET calculated
                    //Note: ... for a short crop having height of 0.12 m (similar to grass), albedo of 0.23, surface resistance of 70 sm−1 for
                    //Note: ... 24-h calculation time steps, and 50 sm−1 for hourly or shorter periods during daytime and 200 sm−1 during"
                    //Note: Jensen and Allen (2015) "Standardized Reference Evapotranspiration for Tall Reference ETrs: Reference ET calculated
                    //Note: ... for a tall crop having height of 0.50m (similar to alfalfa), albedo of 0.23, surface resistance of 45 sm−1 for
                    //Note: ... 24-h calculation time steps, and 30 sm−1 for hourly or shorter periods during daytime and 200 sm−1 during nighttime
                    //Note: ... with daytime G=Rn =0.04 and nighttime G=Rn =0.2."
                    //Note: Short crop coefficients 0.1 during day, 0.5 during night. Tall crop coefficients 0.04 during day, 0.2 during night. 
                    //Note: Jensen et al. (2015) generally suggest using reference crop coefficients and then adjusting to actual conditions
                    //If Radiation_Net_Wpm2 (W/m2) positive, presume during day then GroundHeatFlux_W_p_m2 takes one form
                    if (Radiation_Net_Wpm2 > 0)
                    {
                        //GroundHeatFlux_W_p_m2 (W/m2) defined for tall crop after Jensen et. al., (2015) Chapter 5
                        GroundHeatFlux_W_p_m2 = 0.04 * Radiation_Net_Wpm2;
                    }
                    //Else If Radiation_Net_Wpm2 (W/m2) negative, presume during night then GroundHeatFlux_W_p_m2 takes another form
                    else
                    {
                        //GroundHeatFlux_W_p_m2 (W/m2) defined for tall crop after Jensen et. al., (2015) Chapter 5
                        GroundHeatFlux_W_p_m2 = 0.2 * Radiation_Net_Wpm2;
                    }

                    //LeafAreaIndex_Active_Tree_m2_p_m2 (m2/m2) Eq 13.11 from Chin (2021)
                    //Note: Chin (2021) cites Allen (1998), explaining crops typically have 50% of LAI active; trees may have less
                    LeafAreaIndex_Active_Tree_m2_p_m2 = LeafAreaIndex_list[i].Lai * 0.5;

                    //BulkStomatalResistance_s_p_m (s/m) = 200 based on Shuttleworth (1990) Eq 4.2.22 and interpretation of Allen (1998), Chin (2021) around Eq 13.9
                    //Note: ResistanceSurface_Tree_and_Soil_s_p_m derived from BulkStomatalResistance_s_p_m by division with LAI
                    //Note: Minimum values of ResistanceSurface_Tree_and_Soil_s_p_m ~ 100 s/m in Table 2 of Liang et al. (1994) VIC model and ...
                    //Note: ... Box 5 of Allen et al. (1998) Chapter 2 - FAO Penman-Monteith equation https://www.fao.org/3/x0490e/x0490e06.htm, ...
                    //Note: ... and minimum values of ResistanceSurface_Tree_and_Soil_s_p_m ~ 40 s/m are found in Chpt 11 of Jensen and Allen (2015)
                    ResistanceSurface_Vegetation_s_p_m = 200;

                    //BulkStomatalResistance_Adjustment_frac based on suggestion of Jensen and Allen (2015) Chapter 4 Energy Balance:
                    //Note: ... "In addition, a recommendation by Allen et al. (2006a) to use the same 50 sm−1 surface resistance for hourly
                    //Note: ... or shorter periods during daytime and 200 sm−1 during nighttime with the FAO-56 Penman-Monteith method has
                    //Note: ... made the FAO ETo and ETos references equivalent for hourly time steps and for 24-h periods."
                    //Note: Allen (personal communication) suggests larger stomatal resistance for vegetation in a landscape,
                    //Note: ... i.e., not irrigated reference crop, stating it evoled to survive by restricting water loss. 
                    //Note: Equation derived here and coefficient 0.9 was adjusted to fit expected trends in evapotranspiration 
                    BulkStomatalResistance_Adjustment_frac = Math.Pow((1 - VaporPressure_Deficit_kPa / weatherData[i].SatVPrsKpa), 0.9);
                    ResistanceSurface_Vegetation_s_p_m = ResistanceSurface_Vegetation_s_p_m / BulkStomatalResistance_Adjustment_frac;

                    //If LAI_Tree_m2_p_m2 < 1 then prohibit ResistanceSurface_Tree_and_Soil_s_p_m from going toward infinity with division by LAI < 1
                    if (LeafAreaIndex_list[i].Lai < 1)
                    {
                        ResistanceSurface_Tree_and_Soil_s_p_m = ResistanceSurface_Vegetation_s_p_m / 1.0;
                    }
                    //Else If LAI_Tree_m2_p_m2 >= 1 then ResistanceSurface_Tree_and_Soil_s_p_m divided by LeafAreaIndex_Active_Tree_m2_p_m2 
                    else
                    {
                        //ResistanceSurface_Tree_and_Soil_s_p_m (s/m) defined with Eq 13.9 in Chin (2021) and Eq 4.2.22 in Shuttleworth(1993)
                        ResistanceSurface_Tree_and_Soil_s_p_m = ResistanceSurface_Vegetation_s_p_m / LeafAreaIndex_Active_Tree_m2_p_m2;
                    }
                    //ResistanceSurface_Water_s_p_m defined as 0, which is the case for water surface, defined by Chin (2021) after Eq 13.12
                    ResistanceSurface_Water_s_p_m = 0;

                    //Note: These terms used in HydroPlus to compute evaporation for different cover types
                    //LAI_plus_BAI_plus_GAI_Tree_to_SVeg_ratio is ratio of (LAI + BAI + GAI) for tree to (LAI + BAI + GAI) for short veg
                    //Note: LAI = leaf area index, BAI = bark area index, GAI = ground area index, where GAI = 1 for bare and snow covered ground
                    LAI_plus_BAI_plus_GAI_Tree_to_SVeg_ratio = (LeafAreaIndex_list[i].Lai + BarkAreaIndex_Tree_m2_p_m2 + GroundAreaIndex_m2_p_m2) /
                        (LeafAreaIndex_SVeg_m2_p_m2 + BarkAreaIndex_SVeg_m2_p_m2 + GroundAreaIndex_m2_p_m2);
                    //LAI_plus_BAI_plus_GAI_Ground_to_Tree_ratio is ratio of (LAI + BAI + GAI) for ground to (LAI + BAI + GAI) for tree
                    //Note: (LAI + BAI + GAI) for ground = (0 + 0 + GAI) for ground, despite overhanging vegetation
                    LAI_plus_BAI_plus_GAI_Ground_to_Tree_ratio = (0 + 0 + GroundAreaIndex_m2_p_m2) /
                        (LeafAreaIndex_list[i].Lai + BarkAreaIndex_Tree_m2_p_m2 + GroundAreaIndex_m2_p_m2);

                    //ResistanceAerodynamic_TreeCanopy_s_p_m (s/m) from Eq 13.2 of Chin (2021) or Eq 4.2.25 of Shuttleworth (1993) 
                    //Note: Eq 4.2.25 should use the station windspeed, not a windspeed estimated at the height of any lower object
                    //Note: Height_WeatherStationWindSensor_m is used in place of Height_WeatherStationWindSensor_m to ensure real values from log
                    ResistanceAerodynamic_TreeCanopy_s_p_m = (Math.Log((Height_TreeWindSensor_m - Height_TreeZeroPlaneDisplacement_m) /
                        (rZom * Height_TreeCanopyTop_m)) * Math.Log((Height_TreeWindSensor_m - Height_TreeZeroPlaneDisplacement_m) /
                        (rZov * Height_TreeCanopyTop_m))) / (Math.Pow(vonKarman_Constant, 2) * WindSpeed_Station_m_p_s);

                    //ResistanceAerodynamic_SurfaceWater_s_p_m is Eq 13.8 Chin (2021) or Eq 4.2.29 from Shuttleworth (1993) for open water
                    //Note: RoughnessLength_Water_m = 1.37 mm from Chin (2021)
                    //Note: Height_WeatherStationWindSensor_m can be any height with accompanying WindSpeed_Station_m_p_s
                    ResistanceAerodynamic_SurfaceWater_s_p_m = (4.72 * Math.Pow((Math.Log(Height_WeatherStationWindSensor_m /
                        RoughnessLength_Water_m)), 2)) / (1 + 0.536 * WindSpeed_Station_m_p_s);

                    //ResistanceAerodynamic_Snow_to_Rain_ratio set to 10 according to Lundberg et al. (1998) Eq 6 and 7
                    //Note: ResistanceAerodynamic_Snow_to_Rain_ratio is ratio of aerodynamic resistance for snow to rain 
                    ResistanceAerodynamic_Snow_to_Rain_ratio = 10;

                    //ResistanceAerodynamic_SnowTreeCanopy_s_p_m contains resistance terms in Eq 4 of Fassnacht & Eq 3 of Light (1941)
                    ResistanceAerodynamic_SnowTreeCanopy_s_p_m = (Math.Log(Height_TreeWindSensor_m / RoughnessLength_Snow_m) *
                        Math.Log(Height_TreeWindSensor_m / RoughnessLength_Snow_m)) / (Math.Pow(vonKarman_Constant, 2) * WindSpeed_TreeCanopy_m_p_s);

                    //ResistanceAerodynamic_SnowGround_s_p_m contains resistance terms in Eq 4 of Fassnacht & Eq 3 of Light (1941)
                    //Note: Ground uses wind speed measured at station height
                    ResistanceAerodynamic_SnowGround_s_p_m = (Math.Log(Height_WeatherStationWindSensor_m / RoughnessLength_Snow_m) *
                        Math.Log(Height_WeatherStationWindSensor_m / RoughnessLength_Snow_m)) / (Math.Pow(vonKarman_Constant, 2) * WindSpeed_Station_m_p_s);

                    //PtTrMh (m/h), evapotranspiration from canopy and ground, from Eq Eq 13.1 of Chin (2021), Penman Monteith Evaporation method
                    //Note: Equation 13.1 of Chin (2021) requires all terms are in fundamental SI units (J, Pa, kg, s, m, C, etc.)
                    //Note: Issues with Eq 4.2.27 of Shuttleworth (1993) include not dividing by density of water, usng mixed units MJ/day, kPa/C, etc.
                    weatherData[i].PtTrMh = 1 / (LatentHeatOfVaporizaton_J_p_kg * DensityWater_kg_p_m3) *
                        ((VaporPressureGradient_Pa * (Radiation_Net_Wpm2 - GroundHeatFlux_W_p_m2)) +
                        (DensityAir_kg_p_m3 * SpecificHeat_MoistAir_J_p_kg_p_C * VaporPressure_Deficit_Pa / ResistanceAerodynamic_TreeCanopy_s_p_m)) /
                        (VaporPressureGradient_Pa + PsychrometricConstant_Pa_p_C *
                        (1 + ResistanceSurface_Tree_and_Soil_s_p_m / ResistanceAerodynamic_TreeCanopy_s_p_m));
                    //PtTrMh (m/h) converted from (m/s) with Ratio_sec_to_hr
                    weatherData[i].PtTrMh = weatherData[i].PtTrMh * Ratio_sec_to_hr;
                    //If negative pet to zero
                    if (weatherData[i].PtTrMh < 0) { weatherData[i].PtTrMh = 0.0; }

                    //Note: Adjustment to approximate theory of Jensen and Allen (2015) Chapter 6, Evaporation from Water Surfaces:
                    //Note: ... "An important distinction between Rn for a water body and Rn for vegetation or soil is that with soil and
                    //Note: ... vegetation, essentially all of the Rn quantity is captured at the “opaque” surface and is immediately available
                    //Note: ... solar radiation, Rs, for conversion to λE or H or conduction into the surface as G. 
                    //Note: ... With water, however, much of the penetrates to some depth in the water body, depending on the turbidity
                    //Note: ... of the water, where it is converted to Qt, heat storage." This approach is modified for smaller waters below.
                    //HeatStorage_Water_J_p_m2_p_s (J/m2/s) from Eq 6-14a and 6-14b of Jensen and Allen (2015) modified from large lakes
                    //Note: Large lakes used HeatStorage_Coefficient_a = 0.5 and HeatStorage_Coefficient_b = 0.8; reduced for smaller waters
                    //Note: Large lakes increased loss from long wave radiation at Julian Day = 180; not simulated for smaller waters
                    //Note: Modification includes Heat Storage transfering energy into Ground Heat Flux 
                    double HeatStorage_Coefficient_a = 0.25;
                    //HeatStorage_Coefficient_b reduced from large lake value of 0.8
                    double HeatStorage_Coefficient_b = 0.05;
                    //HeatStorage_Water_J_p_m2_p_s (J/m2/s) heat storage of water from Eq 6-14a and 6-14b of Jensen and Allen (2015) 
                    HeatStorage_Water_J_p_m2_p_s = HeatStorage_Coefficient_a * (weatherData[i].DirRadWm2 + weatherData[i].DifRadWm2) -
                        HeatStorage_Coefficient_b * weatherData[i].UpLwWm2;

                    //PeGrMh (m/h), evaporation from water surface on ground, from Eq Eq 13.1 of Chin (2021), Penman Monteith Evaporation method
                    //Note: Equation 13.1 of Chin (2021) requires all terms are in fundamental SI units (J, Pa, kg, s, m, C, etc.)
                    //Note: Issues with Eq 4.2.27 of Shuttleworth (1993) include not dividing by density of water, usng mixed units MJ/day, kPa/C, etc.
                    //Note: ResistanceSurface_Water_s_p_m set to zero. 
                    weatherData[i].PeGrMh = 1 / (LatentHeatOfVaporizaton_J_p_kg * DensityWater_kg_p_m3) *
                        ((VaporPressureGradient_Pa * (Radiation_Net_Wpm2 - HeatStorage_Water_J_p_m2_p_s)) +
                        (DensityAir_kg_p_m3 * SpecificHeat_MoistAir_J_p_kg_p_C * VaporPressure_Deficit_OverWater_Pa / ResistanceAerodynamic_SurfaceWater_s_p_m)) /
                        (VaporPressureGradient_Pa + PsychrometricConstant_Pa_p_C *
                        (1 + ResistanceSurface_Water_s_p_m / ResistanceAerodynamic_SurfaceWater_s_p_m));
                    //PeGrMh (m/h) converted from (m/s) with Ratio_sec_to_hr
                    weatherData[i].PeGrMh = weatherData[i].PeGrMh * Ratio_sec_to_hr;
                    //If negative pet to zero
                    if (weatherData[i].PeGrMh < 0) { weatherData[i].PeGrMh = 0.0; }

                    //PeTrMh (m/s), evaporation from canopy, set equal to PeGrMh  (m/s)
                    //Note: Variable name contains Mh but it was previously converted to m/s
                    weatherData[i].PeTrMh = weatherData[i].PeGrMh;

                    //PeSnGrMh (m/h), sublimation from ground, from Eq 2 of Lundberg et al. (1998), presented as Eq 13.1 of Chin (2021).
                    //Note: Eq 2 of Lundberg et al. (1998) has no rs/ra term in the denominator, which Eq 13.1 maintains, given surface resistance rs = 0
                    //Note: ResistanceAerodynamic_Snow_to_Rain_ratio term multiplied w to represent Lundberg et al. (1998) 10x factor from Eq 6 to 7
                    weatherData[i].PeSnGrMh = 1 / (LatentHeatOfVaporizaton_J_p_kg * DensityWater_kg_p_m3) *
                        ((VaporPressureGradient_Pa * (Radiation_Net_Wpm2 - HeatStorage_Water_J_p_m2_p_s)) +
                        (DensityAir_kg_p_m3 * SpecificHeat_MoistAir_J_p_kg_p_C * VaporPressure_Deficit_Pa /
                        (ResistanceAerodynamic_SnowGround_s_p_m * ResistanceAerodynamic_Snow_to_Rain_ratio))) /
                        (VaporPressureGradient_Pa + PsychrometricConstant_Pa_p_C *
                        (1 + ResistanceSurface_Water_s_p_m / ResistanceAerodynamic_SnowGround_s_p_m));
                    //PeSnGrMh (m/h) converted from (m/s) with Ratio_sec_to_hr
                    weatherData[i].PeSnGrMh = weatherData[i].PeSnGrMh * Ratio_sec_to_hr;
                    //If negative pet to zero
                    if (weatherData[i].PeSnGrMh < 0) { weatherData[i].PeSnGrMh = 0.0; }

                    //PeSnTrMh (m/h), sublimation from canopy, from Eq 2 of Lundberg et al. (1998), presented as Eq 13.1 of Chin (2021).
                    //Note: Eq 2 of Lundberg et al. (1998) has no rs/ra term in the denominator, which Eq 13.1 maintains, given surface resistance rs = 0
                    //Note: ResistanceAerodynamic_Snow_to_Rain_ratio term multiplied w to represent Lundberg et al. (1998) 10x factor from Eq 6 to 7
                    //Note: Theory: COMET UCAR Snowmelt Processes International Edition states: Snow on vegetation is exposed to ...
                    //Note: ... more wind and sun than snow on the ground and has a higher surface area to mass ratio, thereby ...
                    //Note: ... making it more prone to sublimation and/or melting.
                    weatherData[i].PeSnTrMh = 1 / (LatentHeatOfVaporizaton_J_p_kg * DensityWater_kg_p_m3) *
                        ((VaporPressureGradient_Pa * (Radiation_Net_Wpm2 - HeatStorage_Water_J_p_m2_p_s)) +
                        (DensityAir_kg_p_m3 * SpecificHeat_MoistAir_J_p_kg_p_C * VaporPressure_Deficit_Pa /
                        (ResistanceAerodynamic_SnowTreeCanopy_s_p_m * ResistanceAerodynamic_Snow_to_Rain_ratio))) /
                        (VaporPressureGradient_Pa + PsychrometricConstant_Pa_p_C *
                        (1 + ResistanceSurface_Water_s_p_m / ResistanceAerodynamic_SnowTreeCanopy_s_p_m));
                    //PeSnTrMh (m/h) converted from (m/s) with Ratio_sec_to_hr
                    weatherData[i].PeSnTrMh = weatherData[i].PeSnTrMh * Ratio_sec_to_hr;
                    //If negative pet to zero
                    if (weatherData[i].PeSnTrMh < 0) { weatherData[i].PeSnTrMh = 0.0; }
                }
            }
            catch (Exception)
            {
                throw;
            }
        }
        public static void CalcPrecipInterceptByCanopy(List<SurfaceWeather> wList, List<LeafAreaIndex> LeafAreaIndex_list, int recCnt, VEGTYPE vegType)
        {
            int i;
            double maxSt;       // maximum canopy storage of precipitation
            double c;           // canopy cover fraction
            double potEvMh;     // potential evaporation for tree or ground (for vegType=Tree and Shrub, respectively)
            double canPrcp;     // precipitation fallen and contacted to canopy at the time step
            double freeThf;     // precipitation fallen through canopy without contact (free throughfall) at the time step
            double canDrip;     // drip of water from canopy in the 2nd stage
            double preSt;       // canopy storage of precipitation at the previous time step
            double preEv;       // evaporation from canopy at the previous time step

            for (i = 0; i < recCnt; i++)
            {
                // previous hour condition 
                preSt = (i == 0) ? 0 : wList[i - 1].VegStMh;
                preEv = (i == 0) ? 0 : wList[i - 1].VegEvMh;

                potEvMh = (vegType == VEGTYPE.TREE) ? wList[i].PeTrMh : wList[i].PeGrMh;
                c = (vegType == VEGTYPE.TREE) ? 1 - Math.Exp(-0.7 * LeafAreaIndex_list[i].Lai) : 1 - Math.Exp(-0.3 * LeafAreaIndex_list[i].Lai);

                maxSt = LEAF_STORAGE_M * LeafAreaIndex_list[i].Lai;

                freeThf = wList[i].RainMh * (1 - c);
                canPrcp = wList[i].RainMh - freeThf;

                // storage
                wList[i].VegStMh = preSt + canPrcp - preEv;
                if (wList[i].VegStMh > maxSt)
                {
                    wList[i].VegStMh = maxSt;
                }
                else if (wList[i].VegStMh < 0)
                {
                    wList[i].VegStMh = 0;
                }

                // evaporation
                if (maxSt == 0) // when evergreen % = 0, LAI = 0 in the winter and thus maxSt = 0
                {
                    wList[i].VegEvMh = 0;
                }
                else
                {
                    wList[i].VegEvMh = Math.Pow(wList[i].VegStMh / maxSt, 2.0 / 3.0) * potEvMh;
                    // cap evaporation with storage
                    if (wList[i].VegEvMh > wList[i].VegStMh)
                    {
                        wList[i].VegEvMh = wList[i].VegStMh;
                    }
                }

                // drip
                canDrip = 0;
                if (wList[i].VegStMh < maxSt)                     // 1st stage
                {
                    //                    wList[i].ThrufallMh = freeThf - wList[i].VegEvMh;   // should be this.
                    wList[i].UnderCanThrufallMh = freeThf;                       // but for compatibility with Hydro, use this
                }
                else
                {
                    if (preSt < maxSt)                          // 2nd stage for the first time
                    {
                        canDrip = canPrcp - (maxSt - preSt) - preEv;
                        if (canDrip < 0)
                        {
                            canDrip = 0;
                        }
                    }
                    else                                       // 2nd stage
                    {
                        canDrip = canPrcp - preEv;
                        if (canDrip < 0)
                        {
                            canDrip = 0;
                        }
                    }
                    wList[i].UnderCanThrufallMh = canDrip + freeThf;
                }

                // interception
                // While it's raining hourly interception is calculated from input - output of water
                // (i.e., Rain on canopy - Under canopy throughfall
                if (wList[i].RainInH > 0)
                {
                    wList[i].VegIntcptMh = wList[i].RainMh - wList[i].UnderCanThrufallMh;
                }
                else
                {
                    wList[i].VegIntcptMh = 0;
                }
            }
        }
        public static void CalcPrecipInterceptByUnderCanopyCover(List<SurfaceWeather> wList, int recCnt)
        {
            int i;
            double maxPervSt;       // maximum pervious cover storage of precipitation
            double maxImpervSt;     // maximum impervious cover storage of precipitation
            double potEvMh;         // potential evaporation for ground
            double prePervSt;       // pervious storage of precipitation at the previous time step
            double preImpervSt;     // impervious storage of precipitation at the previous time step
            double prePervEv;       // pervious evaporation at the previous time step
            double preImpervEv;     // impervious evaporation at the previous time step

            for (i = 0; i < recCnt; i++)
            {
                // previous hour condition 
                prePervSt = (i == 0) ? 0 : wList[i - 1].UnderCanPervStMh;
                preImpervSt = (i == 0) ? 0 : wList[i - 1].UnderCanImpervStMh;
                prePervEv = (i == 0) ? 0 : wList[i - 1].UnderCanPervEvMh;
                preImpervEv = (i == 0) ? 0 : wList[i - 1].UnderCanImpervEvMh;

                potEvMh = wList[i].PeGrMh;

                maxPervSt = PERV_STORAGE_M;
                maxImpervSt = IMPERV_STORAGE_M;

                // storage
                //  pervious cover
                wList[i].UnderCanPervStMh = prePervSt + wList[i].UnderCanThrufallMh - prePervEv;
                if (wList[i].UnderCanPervStMh > maxPervSt)
                {
                    wList[i].UnderCanPervStMh = maxPervSt;
                }
                else if (wList[i].UnderCanPervStMh < 0)
                {
                    wList[i].UnderCanPervStMh = 0;
                }
                //  impervious cover
                wList[i].UnderCanImpervStMh = preImpervSt + wList[i].UnderCanThrufallMh - preImpervEv;
                if (wList[i].UnderCanImpervStMh > maxImpervSt)
                {
                    wList[i].UnderCanImpervStMh = maxImpervSt;
                }
                else if (wList[i].UnderCanImpervStMh < 0)
                {
                    wList[i].UnderCanImpervStMh = 0;
                }

                // evaporation
                //  pervious cover
                wList[i].UnderCanPervEvMh = (wList[i].UnderCanPervStMh / maxPervSt) * potEvMh;
                //  cap evaporation with storage
                if (wList[i].UnderCanPervEvMh > wList[i].UnderCanPervStMh)
                {
                    wList[i].UnderCanPervEvMh = wList[i].UnderCanPervStMh;
                }
                //  impervious cover
                wList[i].UnderCanImpervEvMh = (wList[i].UnderCanImpervStMh / maxImpervSt) * potEvMh;
                //  cap evaporation with storage
                if (wList[i].UnderCanImpervEvMh > wList[i].UnderCanImpervStMh)
                {
                    wList[i].UnderCanImpervEvMh = wList[i].UnderCanImpervStMh;
                }

                // runoff
                wList[i].UnderCanImpervRunoffMh = 0;
                if (preImpervSt < maxImpervSt)
                {
                    wList[i].UnderCanImpervRunoffMh = wList[i].UnderCanThrufallMh - (maxImpervSt - preImpervSt) - wList[i].UnderCanImpervEvMh;
                }
                else if (preImpervSt == maxImpervSt)
                {
                    wList[i].UnderCanImpervRunoffMh = wList[i].UnderCanThrufallMh - wList[i].UnderCanImpervEvMh;
                }
                if (wList[i].UnderCanImpervRunoffMh < 0)
                {
                    wList[i].UnderCanImpervRunoffMh = 0;
                }

                // infiltration
                wList[i].UnderCanPervInfilMh = 0;
                if (prePervSt < maxPervSt)
                {
                    wList[i].UnderCanPervInfilMh = wList[i].UnderCanThrufallMh - (maxPervSt - prePervSt) - wList[i].UnderCanPervEvMh;
                }
                else if (prePervSt == maxPervSt)
                {
                    wList[i].UnderCanPervInfilMh = wList[i].UnderCanThrufallMh - wList[i].UnderCanPervEvMh;
                }
                if (wList[i].UnderCanPervInfilMh < 0)
                {
                    wList[i].UnderCanPervInfilMh = 0;
                }
            }
        }
        public static void CalcPrecipInterceptByNoCanopyCover(List<SurfaceWeather> wList, int recCnt)
        {
            int i;
            double maxPervSt;       // maximum pervious cover storage of precipitation
            double maxImpervSt;     // maximum impervious cover storage of precipitation
            double potEvMh;         // potential evaporation for ground
            double prePervSt;       // pervious storage of precipitation at the previous time step
            double preImpervSt;     // impervious storage of precipitation at the previous time step
            double prePervEv;       // pervious evaporation at the previous time step
            double preImpervEv;     // impervious evaporation at the previous time step

            for (i = 0; i < recCnt; i++)
            {
                // previous hour condition 
                prePervSt = (i == 0) ? 0 : wList[i - 1].NoCanPervStMh;
                preImpervSt = (i == 0) ? 0 : wList[i - 1].NoCanImpervStMh;
                prePervEv = (i == 0) ? 0 : wList[i - 1].NoCanPervEvMh;
                preImpervEv = (i == 0) ? 0 : wList[i - 1].NoCanImpervEvMh;

                potEvMh = wList[i].PeGrMh;

                maxPervSt = PERV_STORAGE_M;
                maxImpervSt = IMPERV_STORAGE_M;

                // storage
                //  pervious cover
                wList[i].NoCanPervStMh = prePervSt + wList[i].RainMh - prePervEv;
                if (wList[i].NoCanPervStMh > maxPervSt)
                {
                    wList[i].NoCanPervStMh = maxPervSt;
                }
                else if (wList[i].NoCanPervStMh < 0)
                {
                    wList[i].NoCanPervStMh = 0;
                }
                //  impervious cover
                wList[i].NoCanImpervStMh = preImpervSt + wList[i].RainMh - preImpervEv;
                if (wList[i].NoCanImpervStMh > maxImpervSt)
                {
                    wList[i].NoCanImpervStMh = maxImpervSt;
                }
                else if (wList[i].NoCanImpervStMh < 0)
                {
                    wList[i].NoCanImpervStMh = 0;
                }

                // evaporation
                //  pervious cover
                wList[i].NoCanPervEvMh = (wList[i].NoCanPervStMh / maxPervSt) * potEvMh;
                //  cap evaporation with storage
                if (wList[i].NoCanPervEvMh > wList[i].NoCanPervStMh)
                {
                    wList[i].NoCanPervEvMh = wList[i].NoCanPervStMh;
                }
                //  impervious cover
                wList[i].NoCanImpervEvMh = (wList[i].NoCanImpervStMh / maxImpervSt) * potEvMh;
                //  cap evaporation with storage
                if (wList[i].NoCanImpervEvMh > wList[i].NoCanImpervStMh)
                {
                    wList[i].NoCanImpervEvMh = wList[i].NoCanImpervStMh;
                }

                // runoff
                wList[i].NoCanImpervRunoffMh = 0;
                if (preImpervSt < maxImpervSt)
                {
                    wList[i].NoCanImpervRunoffMh = wList[i].RainMh - (maxImpervSt - preImpervSt) - wList[i].NoCanImpervEvMh;
                }
                else if (preImpervSt == maxImpervSt)
                {
                    wList[i].NoCanImpervRunoffMh = wList[i].RainMh - wList[i].NoCanImpervEvMh;
                }
                if (wList[i].NoCanImpervRunoffMh < 0)
                {
                    wList[i].NoCanImpervRunoffMh = 0;
                }

                // infiltration
                wList[i].NoCanPervInfilMh = 0;
                if (prePervSt < maxPervSt)
                {
                    wList[i].NoCanPervInfilMh = wList[i].RainMh - (maxPervSt - prePervSt) - wList[i].NoCanPervEvMh;
                }
                else if (prePervSt == maxPervSt)
                {
                    wList[i].NoCanPervInfilMh = wList[i].RainMh - wList[i].NoCanPervEvMh;
                }
                if (wList[i].NoCanPervInfilMh < 0)
                {
                    wList[i].NoCanPervInfilMh = 0;
                }
            }
        }
        public static void CopySurfaceWeatherData(string inFile, string outFile, bool blCopy)
        {
            StreamReader sr;
            StreamWriter sw;
            string lineOrg = "";
            string line = "";
            WeatherDataFormat weaFormat;

            using (sr = new StreamReader(inFile))
            {
                weaFormat = CheckWeatherDataFormat(inFile);
                try
                {
                    //First file to be copied: copy the entire input file to the output file.
                    if (blCopy)
                    {
                        using (sw = new StreamWriter(outFile, false))
                        {
                            while ((lineOrg = sr.ReadLine()) != null)
                            {
                                switch (weaFormat)
                                {
                                    case WeatherDataFormat.NEWINTL:
                                        line = lineOrg.Substring(0, 66) + lineOrg.Substring(75, 66);
                                        break;
                                    case WeatherDataFormat.NEWUSCAN:
                                        line = lineOrg.Substring(0, 66) + lineOrg.Substring(81, 66);
                                        break;
                                    case WeatherDataFormat.OLD:
                                        line = lineOrg;
                                        break;
                                    case WeatherDataFormat.NARCCAP:
                                        line = lineOrg;
                                        break;
                                    default:
                                        break;
                                }
                                sw.WriteLine(line);
                            }
                        }
                    }
                    //Append the file by excluding the first header line.
                    else
                    {
                        using (sw = new StreamWriter(outFile, true))
                        {
                            //Read the header line
                            lineOrg = sr.ReadLine();
                            while ((lineOrg = sr.ReadLine()) != null)
                            {
                                switch (weaFormat)
                                {
                                    case WeatherDataFormat.NEWINTL:
                                        line = lineOrg.Substring(0, 66) + lineOrg.Substring(75, 66);
                                        break;
                                    case WeatherDataFormat.NEWUSCAN:
                                        line = lineOrg.Substring(0, 66) + lineOrg.Substring(81, 66);
                                        break;
                                    case WeatherDataFormat.OLD:
                                        line = lineOrg;
                                        break;
                                    default:
                                        break;
                                }
                                sw.WriteLine(line);
                            }
                        }
                    }
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }
        public static void WriteSurfaceWeatherRecords(string WeatherVariables_dataBase, List<SurfaceWeather> WeatherVariables_list, int recCnt)
        {
            try
            {
                string[] weaStrings = new string[WeatherVariables_list.Count + 1]; //String array to be written to Meteo.csv
                weaStrings[0] = "TimeStamp,DownLwWm2,UpLwWm2,LatRad,HourAngleRad,DecAngleRad,AirDens_kg_p_m3," +
                    "AirMass,Ceiling,ToCldCov,OpCldCov,TrCldCov,PrsIn,PrsKPa,PrsMBar,PeTrMh,PeGrMh," +
                    "PeSnTrMh,PeSnGrMh,PtTrMh,VegEvMh,VegStMh,VegIntcptMh,UnderCanThrufallMh," +
                    "UnderCanPervEvMh,UnderCanPervStMh,UnderCanPervInfilMh,UnderCanImpervEvMh," +
                    "UnderCanImpervStMh,UnderCanImpervRunoffMh,NoCanPervEvMh,NoCanPervStMh," +
                    "NoCanPervInfilMh,NoCanImpervEvMh,NoCanImpervStMh,NoCanImpervRunoffMh,DirRadWm2," +
                    "DifRadWm2,GlbRadWm2,PARWm2,PARuEm2s,NetRadWm2,RainInH,RainMh,RelHum,SatVPrsKPa," +
                    "SnowIn,SnowM,SolZenAgl,TempC,TempF,TempK,DewTempC,DewTempF,VapPrsKPa,WdDir,WdSpdMh," +
                    "WdSpdMs,WdSpdKnt"; //CSV header
                for (int i = 0; i < WeatherVariables_list.Count; i++)
                {
                    string[] rowString = new string[59]; //String array for new row in Meteo.csv
                    rowString[0] = Convert.ToString(WeatherVariables_list[i].TimeStamp);
                    rowString[1] = Convert.ToString(WeatherVariables_list[i].DownLwWm2);
                    rowString[2] = Convert.ToString(WeatherVariables_list[i].UpLwWm2);
                    rowString[3] = Convert.ToString(WeatherVariables_list[i].LatRad);
                    rowString[4] = Convert.ToString(WeatherVariables_list[i].HourAngleRad);
                    rowString[5] = Convert.ToString(WeatherVariables_list[i].DecAngleRad);
                    rowString[6] = Convert.ToString(WeatherVariables_list[i].AirDens_kg_p_m3);
                    rowString[7] = Convert.ToString(WeatherVariables_list[i].AirMass);
                    rowString[8] = Convert.ToString(WeatherVariables_list[i].Ceiling);
                    rowString[9] = Convert.ToString(WeatherVariables_list[i].ToCldCov);
                    rowString[10] = Convert.ToString(WeatherVariables_list[i].OpCldCov);
                    rowString[11] = Convert.ToString(WeatherVariables_list[i].TrCldCov);
                    rowString[12] = Convert.ToString(WeatherVariables_list[i].PrsIn);
                    rowString[13] = Convert.ToString(WeatherVariables_list[i].PrsKpa);
                    rowString[14] = Convert.ToString(WeatherVariables_list[i].PrsMbar);
                    rowString[15] = Convert.ToString(WeatherVariables_list[i].PeTrMh);
                    rowString[16] = Convert.ToString(WeatherVariables_list[i].PeGrMh);
                    rowString[17] = Convert.ToString(WeatherVariables_list[i].PeSnTrMh);
                    rowString[18] = Convert.ToString(WeatherVariables_list[i].PeSnGrMh);
                    rowString[19] = Convert.ToString(WeatherVariables_list[i].PtTrMh);
                    rowString[20] = Convert.ToString(WeatherVariables_list[i].VegEvMh);
                    rowString[21] = Convert.ToString(WeatherVariables_list[i].VegStMh);
                    rowString[22] = Convert.ToString(WeatherVariables_list[i].VegIntcptMh);
                    rowString[23] = Convert.ToString(WeatherVariables_list[i].UnderCanThrufallMh);
                    rowString[24] = Convert.ToString(WeatherVariables_list[i].UnderCanPervEvMh);
                    rowString[25] = Convert.ToString(WeatherVariables_list[i].UnderCanPervStMh);
                    rowString[26] = Convert.ToString(WeatherVariables_list[i].UnderCanPervInfilMh);
                    rowString[27] = Convert.ToString(WeatherVariables_list[i].UnderCanImpervEvMh);
                    rowString[28] = Convert.ToString(WeatherVariables_list[i].UnderCanImpervStMh);
                    rowString[29] = Convert.ToString(WeatherVariables_list[i].UnderCanImpervRunoffMh);
                    rowString[30] = Convert.ToString(WeatherVariables_list[i].NoCanPervEvMh);
                    rowString[31] = Convert.ToString(WeatherVariables_list[i].NoCanPervStMh);
                    rowString[32] = Convert.ToString(WeatherVariables_list[i].NoCanPervInfilMh);
                    rowString[33] = Convert.ToString(WeatherVariables_list[i].NoCanImpervEvMh);
                    rowString[34] = Convert.ToString(WeatherVariables_list[i].NoCanImpervStMh);
                    rowString[35] = Convert.ToString(WeatherVariables_list[i].NoCanImpervRunoffMh);
                    rowString[36] = Convert.ToString(WeatherVariables_list[i].DirRadWm2);
                    rowString[37] = Convert.ToString(WeatherVariables_list[i].DifRadWm2);
                    rowString[38] = Convert.ToString(WeatherVariables_list[i].GlbRadWm2);
                    rowString[39] = Convert.ToString(WeatherVariables_list[i].PARWm2);
                    rowString[40] = Convert.ToString(WeatherVariables_list[i].PARuEm2s);
                    rowString[41] = Convert.ToString(WeatherVariables_list[i].NetRadWm2);
                    rowString[42] = Convert.ToString(WeatherVariables_list[i].RainInH);
                    rowString[43] = Convert.ToString(WeatherVariables_list[i].RainMh);
                    rowString[44] = Convert.ToString(WeatherVariables_list[i].RelHum);
                    rowString[45] = Convert.ToString(WeatherVariables_list[i].SatVPrsKpa);
                    rowString[46] = Convert.ToString(WeatherVariables_list[i].SnowIn);
                    rowString[47] = Convert.ToString(WeatherVariables_list[i].SnowM);
                    rowString[48] = Convert.ToString(WeatherVariables_list[i].SolZenAgl);
                    rowString[49] = Convert.ToString(WeatherVariables_list[i].TempC);
                    rowString[50] = Convert.ToString(WeatherVariables_list[i].TempF);
                    rowString[51] = Convert.ToString(WeatherVariables_list[i].TempK);
                    rowString[52] = Convert.ToString(WeatherVariables_list[i].DewTempC);
                    rowString[53] = Convert.ToString(WeatherVariables_list[i].DewTempF);
                    rowString[54] = Convert.ToString(WeatherVariables_list[i].VapPrsKpa);
                    rowString[55] = Convert.ToString(WeatherVariables_list[i].WdDir);
                    rowString[56] = Convert.ToString(WeatherVariables_list[i].WdSpdMh);
                    rowString[57] = Convert.ToString(WeatherVariables_list[i].WdSpdMs);
                    rowString[58] = Convert.ToString(WeatherVariables_list[i].WdSpdKnt);
                    weaStrings[i + 1] = String.Join(",", rowString); //Join the segments of the new row, then add new row to weaStrings
                }
                string csv = String.Join("\r\n", weaStrings); //Join all CSV rows into one string, with comma + newline as separator
                File.WriteAllText(WeatherVariables_dataBase, csv);
            }
            catch (Exception)
            {
                throw;
            }
        }
        public static int ReadSurfaceAllRecords(string WeatherVariables_dataBase, ref List<SurfaceWeather> sfcData)
        {
            string line;
            SurfaceWeather data;

            using (StreamReader sr = new StreamReader(WeatherVariables_dataBase))
            {
                try
                {
                    //read the header line
                    sr.ReadLine();

                    while ((line = sr.ReadLine()) != null)
                    {
                        data = new SurfaceWeather();
                        data.TimeStamp = DateTime.Parse(line.Split(',')[0]);
                        data.DownLwWm2 = Double.Parse(line.Split(',')[1]);
                        data.UpLwWm2 = Double.Parse(line.Split(',')[2]);
                        data.LatRad = Double.Parse(line.Split(',')[3]);
                        data.HourAngleRad = Double.Parse(line.Split(',')[4]);
                        data.DecAngleRad = Double.Parse(line.Split(',')[5]);
                        data.AirDens_kg_p_m3 = Double.Parse(line.Split(',')[6]);
                        data.AirMass = Double.Parse(line.Split(',')[7]);
                        data.Ceiling = Double.Parse(line.Split(',')[8]);
                        data.ToCldCov = Double.Parse(line.Split(',')[9]);
                        data.OpCldCov = Double.Parse(line.Split(',')[10]);
                        data.TrCldCov = Double.Parse(line.Split(',')[11]);
                        data.PrsIn = Double.Parse(line.Split(',')[12]);
                        data.PrsKpa = Double.Parse(line.Split(',')[13]);
                        data.PrsMbar = Double.Parse(line.Split(',')[14]);
                        data.PeTrMh = Double.Parse(line.Split(',')[15]);
                        data.PeGrMh = Double.Parse(line.Split(',')[16]);
                        data.PeSnTrMh = Double.Parse(line.Split(',')[17]);
                        data.PeSnGrMh = Double.Parse(line.Split(',')[18]);
                        data.PtTrMh = Double.Parse(line.Split(',')[19]);
                        data.VegEvMh = Double.Parse(line.Split(',')[20]);
                        data.VegStMh = Double.Parse(line.Split(',')[21]);
                        data.VegIntcptMh = Double.Parse(line.Split(',')[22]);
                        data.UnderCanThrufallMh = Double.Parse(line.Split(',')[23]);
                        data.UnderCanPervEvMh = Double.Parse(line.Split(',')[24]);
                        data.UnderCanPervStMh = Double.Parse(line.Split(',')[25]);
                        data.UnderCanPervInfilMh = Double.Parse(line.Split(',')[26]);
                        data.UnderCanImpervEvMh = Double.Parse(line.Split(',')[27]);
                        data.UnderCanImpervStMh = Double.Parse(line.Split(',')[28]);
                        data.UnderCanImpervRunoffMh = Double.Parse(line.Split(',')[29]);
                        data.NoCanPervEvMh = Double.Parse(line.Split(',')[30]);
                        data.NoCanPervStMh = Double.Parse(line.Split(',')[31]);
                        data.NoCanPervInfilMh = Double.Parse(line.Split(',')[32]);
                        data.NoCanImpervEvMh = Double.Parse(line.Split(',')[33]);
                        data.NoCanImpervStMh = Double.Parse(line.Split(',')[34]);
                        data.NoCanImpervRunoffMh = Double.Parse(line.Split(',')[35]);
                        data.DirRadWm2 = Double.Parse(line.Split(',')[36]);
                        data.DifRadWm2 = Double.Parse(line.Split(',')[37]);
                        data.GlbRadWm2 = Double.Parse(line.Split(',')[38]);
                        data.PARWm2 = Double.Parse(line.Split(',')[39]);
                        data.PARuEm2s = Double.Parse(line.Split(',')[40]);
                        data.NetRadWm2 = Double.Parse(line.Split(',')[41]);
                        data.RainInH = Double.Parse(line.Split(',')[42]);
                        data.RainMh = Double.Parse(line.Split(',')[43]);
                        data.RelHum = Double.Parse(line.Split(',')[44]);
                        data.SatVPrsKpa = Double.Parse(line.Split(',')[45]);
                        data.SnowIn = Double.Parse(line.Split(',')[46]);
                        data.SnowM = Double.Parse(line.Split(',')[47]);
                        data.SolZenAgl = Double.Parse(line.Split(',')[48]);
                        data.TempC = Double.Parse(line.Split(',')[49]);
                        data.TempF = Double.Parse(line.Split(',')[50]);
                        data.TempK = Double.Parse(line.Split(',')[51]);
                        data.DewTempC = Double.Parse(line.Split(',')[52]);
                        data.DewTempF = Double.Parse(line.Split(',')[53]);
                        data.VapPrsKpa = Double.Parse(line.Split(',')[54]);
                        data.WdDir = Double.Parse(line.Split(',')[55]);
                        data.WdSpdMh = Double.Parse(line.Split(',')[56]);
                        data.WdSpdMs = Double.Parse(line.Split(',')[57]);
                        data.WdSpdKnt = Double.Parse(line.Split(',')[58]);
                        sfcData.Add(data);
                    }
                    return (sfcData.Count);
                }
                catch (Exception)
                {
                    throw;
                }
            }
        }
    }
}
