#z_PrecipitationData_GIOVANNI.py will process NASA GIOVANNI server GPM_3IMERGHH data into format usable by i-Tree
#Note: Script can be run in two approaches, explained below as A and B:

#Created by Jay Heppler at Davey Institute for processing of international weather data lacking precipitation information for i-Tree's Research Suite. 
#Updated by T. Endreny at SUNY ESF: created query to user with command prompt to determine whether they will generate only Precipitation.csv or Weather.csv; changed precipitation formatting to include date; corrected summing two 30 min depths; changed start row 11 to 9; reformat to .csv inputs

#Approach A
#Step 1a: Download WeatherPrep scripts and utility from i-Tree Tools and save as: C:\iTree\WeatherPrep\
#Step 1b: Follow directions in C:\iTree\WeatherPrep\weather_data_NOAA_tool\ReadMe.txt to obtain raw NOAA data 
#Step 1c: Create work folder such as C:\iTree\a_prep_Weather\syr_ny\2022\ to contain the raw NOAA data
#Step 2a: Browse GIOVANNI data at https://giovanni.gsfc.nasa.gov/giovanni/ using keyword GPM_3IMERGHH for AOI
#Step 2b: Download GPM_3IMERGHH product as Time Series, Area Averaged to work folder created above
#Step 2c: Keep original name of GPM_3IMERGHH product, e.g., g4.areaAvgTimeSeries.GPM_3IMERGHH_06_precipitationCal..
#Step 3a: Run this GIOVANNI script and select data output option letter P, for Precipitation.csv
#Step 4a: Follow directions in C:\iTree\WeatherPrep\TestCases\ReadMe.txt
#Step 4b: Modify WeatherPrepConfig.xml element PrecipitationDataCsv to point to output from this script
#Step 4c: Generate weather.csv file that contains the GIOVANNI precipitation data

#Approach B
#Step 1a: Download WeatherPrep scripts and utility from i-Tree Tools and save as: C:\iTree\WeatherPrep\
#Step 1b: Follow directions in C:\iTree\WeatherPrep\weather_data_NOAA_tool\ReadMe.txt to obtain raw NOAA data 
#Step 1c: Create work folder such as C:\iTree\a_prep_Weather\syr_ny\2022\ to contain the raw NOAA data
#Step 2a: Follow directions in C:\iTree\WeatherPrep\TestCases\ReadMe.txt
#Step 2b: Modify WeatherPrepConfig.xml element PrecipitationDataCsv to be blank
#Step 2c: Generate 1st estimate of weather.csv file that does not yet contain the GIOVANNI precipitation data
#Step 3a: Browse GIOVANNI data at https://giovanni.gsfc.nasa.gov/giovanni/ using keyword GPM_3IMERGHH for AOI
#Step 3b: Download GPM_3IMERGHH product as Time Series, Area Averaged to work folder created above
#Step 3c: Keep original name of GPM_3IMERGHH product, e.g., g4.areaAvgTimeSeries.GPM_3IMERGHH_06_precipitationCal..
#Step 4a: Run this GIOVANNI script and select data output option letter W, for Weather.csv
#Step 4b: Generate weather.csv file that contains the GIOVANNI precipitation data

#Precipitation.csv formatting is:
#Date_YYYYMMDD,Time_HH:MM:SS,Precipitation_m_p_h(m/h)
#20190101,00:00:00,0.00076200
#20190101,01:00:00,0.00000000

import csv
import os 
import datetime 
import glob 
import shutil
import pandas as pd

#Before running script, put weather file of interest and GIOVANNI file of interest into same folder. 
os.chdir(os.path.dirname(__file__))

#precip_find string variable for path of Giovanni IMERGE input defined as GIOVANNI file using key word search *GPM_3IMERGHH* 
#Note: glob generates a list, [0] pulls out the first item of this list as a string
precip_find = glob.glob('*GPM_3IMERGHH*')

#if precip_find returns no results
if len(list(precip_find)) == 0:
    print("GIOVANNI file type should contain keyword 'GPM_3IMERGHH'. Ensure you have a file with this keyword in directory", os.getcwd())
else:
    pass 
    
#Prompt for command line input whether program generates Precipitation.csv or Weather.csv
output_choice = input('Generate Precipitation.csv or Weather.csv (P or W):')
#Substring output_choice to ensure accurate input
output_choice = output_choice[0:]

#Initiate flags
flag_Precipitation = 0
flag_Weather = 0

#Define 1st row of Giovanni IMERGE precipitation data
row_start_precipitation_input = 9

#If output_choice is P or p then only generate Precipitation.csv file
if (output_choice == 'P' or output_choice == 'p'):
    #Confirm output file
    print("Program will create Precipitation.csv")
    #Set flag_Precipitation to 1
    flag_Precipitation = 1
    #set flag_Weather to 0
    flag_Weather = 0
else:
    #Prompt for command line input whether program generates Precipitation.csv or Weather.csv
    input_choice = input('Generate Weather.csv for model i-Tree Energy or i-Tree Hydro (E or H):')
    #Substring input_choice to ensure accurate input
    input_choice = input_choice[0:]
    #Confirm output file
    print("Program will create Weather.csv")
    #set flag_Weather to 1
    flag_Weather = 1
    #Set flag_Precipitation to 0
    flag_Precipitation = 0

#precip_file defined by Python function os.path.join with 1st part of precip_find string variable
precip_file = os.path.join(os.getcwd(), str(precip_find[0]))

#Create Precipitation.csv as time series of Giovanni IMERGE precipitation 
precip_file_copy = os.path.join(os.getcwd(), 'Giovanni_IMERGE_Data.csv')
#Create working copies using shutil function
shutil.copy(precip_file, precip_file_copy)

if flag_Weather == 1:
    #find Weather file, create variable containing full path name. 
    weather_find = glob.glob('Weather.csv')  
    weather_file = os.path.join(os.getcwd(), str(weather_find[0]))
    #Create working copy directories
    weather_file_copy = os.path.join(os.getcwd(), 'Weather.wc.csv')
    #Create working copies
    shutil.copy(weather_file, weather_file_copy)

#with open Python command uses precip_file_copy, named f0
with open(precip_file_copy) as f0:
    #lines_precip holds all f0 lines using Python csv.reader function
    lines_precip = csv.reader(f0)
    #raw_data_precip holds all lines_precip as rows
    raw_data_precip = [row for row in lines_precip]
    
    #row_count_p variable defines length of input file, raw_data_precip, to define length of loop
    row_count_p = len(list(raw_data_precip))
    
    #first_datetime defined using Python function datetime.datetime.strptime 
    #Note: "%Y-%m-%d %H:%M:%S" defines format of Giovanni IMERGE input date and time
    first_datetime = datetime.datetime.strptime(raw_data_precip[row_start_precipitation_input][0], "%Y-%m-%d %H:%M:%S")
        
    #If first_datetime.minute greater than 1 then it is row at a 30 minute time interval, so advance 1 row
    if first_datetime.minute >= 1:
        first_datetime = datetime.datetime.strptime(raw_data_precip[row_start_precipitation_input + 1][0], "%Y-%m-%d %H:%M:%S")
        start_range = row_start_precipitation_input + 1
    else:
        start_range = row_start_precipitation_input
        
    #Confirm start range
    print('First line of Giovanni IMERGE data at row {}: {}'.format(start_range, first_datetime))
    print("If you disagree, open script and change value of row_start_precipitation_input.")

    #create empty list for CSV
    output_data_precip = []
    
    #for loop at interval of 2, reading in value at each hour, accounting for half hour time steps 
    for j in range(start_range, row_count_p, 2):
        #precip_mm is sum of two 30 minute values, giving the 60 minute total
        precip_mm = (float(raw_data_precip[j][1]) + float(raw_data_precip[j+1][1]))
    
        #cte Should precipitation total for two 30 min periods be summed to get precipitation total for 60 min, not averaged
        #find average between the two observations of hour averages
        #precip_mm = ((float(raw_data_precip[j][1])+float(raw_data_precip[j+1][1]))/2)
        
        #precip_m defined as m/h, converted from from mm/h by division of 1000
        precip_m = (float(precip_mm)/1000) 
        #row_count_precip equals precip_m
        row_data_precip = [precip_m]
        #add precipitation value to list
        output_data_precip.append(row_data_precip)
        
        #df_p is created dataframe to store values
        df_p = pd.DataFrame(output_data_precip)
        
        #create intermediate CSV to store precipitation values 
        output_file_precip = os.path.join(os.getcwd(), 'Precipitation.csv')
        df_p.to_csv(output_file_precip, header= False, index= False)
        
#If flag_Weather is true then write preciptation to Weather.csv
if flag_Weather == 1:
    #open weather file 
    with open(weather_file_copy) as f1:
        lines_weather = csv.reader(f1)
        raw_data_weather = [row for row in lines_weather]
        #get row count to determine length of loop
        row_count_weather = len(list(raw_data_weather))
        row_count_precip = len(output_data_precip)
        print(row_count_precip)
        
        #Warning in case row_count_weather and row_count_precip do not match. If the weather is an annual generation, this may be because the weather data for December 31 is missing. 
        if row_count_weather == row_count_precip:
            print("Number of weather and precipitation observations match. Writing final weather file now.")
        if row_count_weather > row_count_precip: 
            print("There are more weather observations than precipitation observations. This script will only write the maximum number of weather observations. Please ensure that your start times and dates match to avoid corruption of data. Writing final weather file now.")
        if row_count_weather < row_count_precip:
            print("There are more precipitation observations than weather observations. This script will only write the maximum number of weather observations. Please ensure that your start times and dates match to avoid corruption of data. Writing final weather file now.")

        #If input_choice is E or e then generate Weather.csv with WindDir_deg variable
        if (input_choice== 'E' or input_choice == 'e'):
            #header for Weather file 
            output_data_weather = [["YYYYMMDD","HH:MM:SS", "Tair_F(F)","Tdew_F(F)","NetRad_Wpm2(W/m^2)","WndSpd_mps(m/s)","WindDir_deg(deg)","AtmPres_kPa(kPa)","Precip_mph(m/h)","SnowDepth_m(m)"]]
        #If input_choice is not E or e then generate Weather.csv without WindDir_deg variable
        else:
            #header for Weather file 
            output_data_weather = [["YYYYMMDD","HH:MM:SS", "Tair_F(F)","Tdew_F(F)","NetRad_Wpm2(W/m^2)","WndSpd_mps(m/s)","AtmPres_kPa(kPa)","Precip_mph(m/h)","SnowDepth_m(m)"]]        
        
        for j in range(1, row_count_weather, 1):
            #If input_choice is E or e then generate Weather.csv with WindDir_deg variable
            if (input_choice== 'E' or input_choice == 'e'):
                YYYYMMDD = raw_data_weather[j][0]
                HHMMSS = raw_data_weather[j][1]
                Tair_F = raw_data_weather[j][2]
                Tdew_F = raw_data_weather[j][3]
                NetRad_Wpm2 = raw_data_weather[j][4]
                WndSpd_mps = raw_data_weather[j][5]
                WindDir_deg = raw_data_weather[j][6]
                AtmPres_kPa = raw_data_weather[j][7]
                Precip_mph = str(output_data_precip[j-1])
                SnowDepth_m = raw_data_weather[j][9]
                #Precip_mph includes brackets [] because of it's call from a list. [1:-1] removes the brackets from the string. 
                row_data_weather = [YYYYMMDD, HHMMSS, Tair_F, Tdew_F, NetRad_Wpm2, WndSpd_mps, WindDir_deg, AtmPres_kPa, Precip_mph[1:-1], SnowDepth_m]
            #If input_choice is not E or e then generate Weather.csv without WindDir_deg variable
            else:
                YYYYMMDD = raw_data_weather[j][0]
                HHMMSS = raw_data_weather[j][1]
                Tair_F = raw_data_weather[j][2]
                Tdew_F = raw_data_weather[j][3]
                NetRad_Wpm2 = raw_data_weather[j][4]
                WndSpd_mps = raw_data_weather[j][5]
                AtmPres_kPa = raw_data_weather[j][6]
                Precip_mph = str(output_data_precip[j-1])
                SnowDepth_m = raw_data_weather[j][8]
                #Precip_mph includes brackets [] because of it's call from a list. [1:-1] removes the brackets from the string. 
                row_data_weather = [YYYYMMDD, HHMMSS, Tair_F, Tdew_F, NetRad_Wpm2, WndSpd_mps, AtmPres_kPa, Precip_mph[1:-1], SnowDepth_m]
            output_data_weather.append(row_data_weather)    
            df_w = pd.DataFrame(output_data_weather)
            #create csv, ultimately will become final Weather.dat. 
            output_file_weather = os.path.join(os.getcwd(), 'Weather2.csv')
            df_w.to_csv(output_file_weather, header= False, index= False)

#If flag_Weather is true then remove and copy certain files
if flag_Weather == 1:        
    os.remove(weather_file_copy)
    #copies new CSV (Weather2) to the same name as the old weather_file
    shutil.copy(output_file_weather, weather_file)
    os.remove(output_file_weather)
    os.remove(output_file_precip)
    print(weather_file, "now contains updated precipitation information.")
else:
    print(output_file_precip, "created to use as PrecipitationDataCsv in WeatherPrepConfig.xml.")

#delete all intermediate files 
os.remove(precip_file_copy)
